#  This work is based on original code developed and copyrighted by TNO 2023.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO
from dataclasses import dataclass
from datetime import datetime
from uuid import uuid4

import pytz
from pyecore.ecore import EDate

import esdl
from esdl.profiles.profilemanager import ProfileManager, UnknownProfileNameException, ProfileType
from influxdb import InfluxDBClient

from esdl.utils.datetime_utils import parse_date


@dataclass
class ConnectionSettings:
    """
    The ConnectionSettings class represents all required settings that are required to connect to an
    InfluxDB server
    """
    host: str
    """hostname of the InfluxDB server to connect to"""
    port: int
    """port to connect to"""
    username: str
    """username of the user that has permission to access the database"""
    password: str
    """password of the user"""
    database: str
    """database to connect to"""
    ssl: bool
    """indicates whether or not to use SSL when connecting to the InfluxDB server"""
    verify_ssl: bool
    """indicates wether or not to verify SSL certificates"""


class InfluxDBProfileManager(ProfileManager):
    """
    InfluxDBProfileManager: manages profile data that can be loaded from and saved to InfluxDB (v1 only).
    InfluxDBProfileManager is a subclass of ProfileManager, so it also provides functionality to convert from/to
    different ESDL profiles and to load/save to CSV
    """
    def __init__(self, settings: ConnectionSettings, source_profile=None):
        """
        Constructor of the InfluxDBProfileManager

        :param settings: connection settings to connect to the InfluxDB server
        :param source_profile: the source profile of which the data is copied into this profiles manager instance
        """
        super().__init__()

        # Temporary fix, for ESDL profiles that have HTTP(S) in the hostname
        settings.host = settings.host.replace("http://", "").replace("https://", "")

        self.database_settings = settings
        self.influxdb_client = InfluxDBClient(
            host=settings.host,
            port=settings.port,
            username=settings.username,
            password=settings.password,
            database=settings.database,
            ssl=settings.ssl,
            verify_ssl=settings.verify_ssl
        )

        if settings.database not in [db['name'] for db in self.influxdb_client.get_list_database()]:
            self.influxdb_client.create_database(settings.database)

        if source_profile:
            self.convert(source_profile)
            if isinstance(source_profile, InfluxDBProfileManager):
                self.database_settings = source_profile.database_settings

    def load_influxdb(self, measurement: str, fields: list, from_datetime: datetime = None, to_datetime: datetime = None, filters: list = None):
        """
        Loads profile information from InfluxDB

        :param measurement: the name of the measurement to use in InfluxDB
        :param fields: a list of field names that need to be loaded from InfluxDB
        :param from_datetime: the start datetime (included in the data)
        :param to_datetime: the end datetime (included in the data)
        :param filters: a list of dictionaries with 'tag' and 'value' keys, that can be used to filter on the data
        :return: None
        """
        self.clear_profile()

        where_clause_list = list()
        if from_datetime and to_datetime:
            influxdb_startdate = from_datetime.strftime('%Y-%m-%dT%H:%M:%SZ')
            influxdb_enddate = to_datetime.strftime('%Y-%m-%dT%H:%M:%SZ')

            where_clause_list.append("time >= '" + influxdb_startdate + "'")
            where_clause_list.append("time <= '" + influxdb_enddate + "'")

        if filters:
            for filter in filters:
                where_clause_list.append(f"{filter['tag']}='{filter['value']}'")

        fields_string = ','.join(['"'+field+'"' for field in fields])
        query_string = f"SELECT {fields_string} FROM \"{measurement}\""
        if where_clause_list:
            query_string += " WHERE (" + " AND ".join(where_clause_list) + ")"

        res = self.influxdb_client.query(query_string)

        if res:
            res_list = res.get_points()
            for elem in res_list:
                dt = parse_date(elem['time'])

                try:
                    aware_dt = pytz.utc.localize(dt)  # Assume timezone is UTC if no TZ was given
                except ValueError:  # ValueError: No naive datetime (tzinfo is already set)
                    aware_dt = dt

                row = [aware_dt]
                if self.start_datetime is None:
                    self.start_datetime = aware_dt

                for f in fields:
                    row.append(elem[f])

                self.profile_data_list.append(row)
                self.num_profile_items += 1

            self.end_datetime = self.profile_data_list[-1][0]
            header = ['datetime']
            header.extend(fields)
            self.profile_header = header
            self.profile_type = ProfileType.DATETIME_LIST

            tags = None
            if filters:
                tags = dict()
                for filter in filters:
                    tags[filter['tag']] = filter['value']
            return self.get_esdl_influxdb_profile(measurement, fields, tags)
        else:
            return None

    @staticmethod
    def _parse_esdl_profile_filters(esdl_filters):
        if esdl_filters is None or esdl_filters == "":
            return []

        filter_str_list = esdl_filters.split(" AND ")
        filter_list = list()
        for f in filter_str_list:
            try:
                [key, value] = f.split("=")
                if key is not None and key != "" and value is not None and value != "":
                    filter_list.append({"key": key, "value": value})
                else:
                    raise WrongFilterFormatException("Filter specification in ESDL profile cannot be parsed")
            except:
                raise WrongFilterFormatException("Filter specification in ESDL profile cannot be parsed")

        return filter_list

    @staticmethod
    def create_esdl_influxdb_profile_manager(
            esdl_profile: esdl.InfluxDBProfile,
            username: str = None,
            password: str = None,
            use_ssl: bool = False,
            verify_ssl: bool = False):
        """
        method to create an instance of InfluxDBProfileManager using an esdl.InfluxDBProfile. The data referenced to in
        the profile is loaded from InfluxDB

        :param esdl_profile: esdl.InfluxDBProfile that is used as an input profile
        :param username: username to be used to connect to the InfluxDB server mentioned in the esdl.InfluxDBProfile,
                         defaults to None
        :param password: password to be used to connect to the InfluxDB server mentioned in the esdl.InfluxDBProfile,
                         defaults to None
        :param use_ssl: indicates if HTTPS should be used instead of HTTP to connect to the InfluxDB server, defaults
                        to False
        :param verify_ssl: verify SSL certificates for HTTPS requests, defaults to False
        """
        conn_settings = ConnectionSettings(
            host=esdl_profile.host,
            port=esdl_profile.port,
            database=esdl_profile.database,
            username=username,
            password=password,
            ssl=use_ssl,
            verify_ssl=verify_ssl
        )

        prof = InfluxDBProfileManager(conn_settings)
        prof.load_influxdb(
            measurement=esdl_profile.measurement,
            fields=[esdl_profile.field],
            from_datetime=esdl_profile.startDate,
            to_datetime=esdl_profile.endDate,
            filters=InfluxDBProfileManager._parse_esdl_profile_filters(esdl_profile.filters),
        )

        return prof

    @staticmethod
    def _create_esdl_filters_from_tags_dict(tags: dict = None):
        if not tags:
            return ""
        else:
            filter_list = list()

            for k, v in tags.items():
                filter_list.append(f"{k}='{v}'")

            return " AND ".join(filter_list)

    def get_esdl_influxdb_profile(self, measurement: str, field_names: list, tags: dict = None):
        """
        Creates an esdl.InfluxDBProfile instance (or a list of instances) that refers to the data in the database.
        Is called by load_influxdb and save_influxdb

        :param measurement: name of the InfluxDB measurement where the data must be written to
        :param field_names: list of the fields that should be written to InfluxDB
        :param tags: dictionary with tags and tag values, that should be used when writing this data to InfluxDB
        :return: an esdl.InfluxDBProfile instance or a list of esdl.InfluxDBProfile instances in case multiple fields
                 were specified, with proper references to the data in the database
        """
        profile_list = list()
        for field in field_names:
            esdl_profile = esdl.InfluxDBProfile(
                id=str(uuid4()),
                name=field,
                host=self.database_settings.host,
                port=self.database_settings.port,
                database=self.database_settings.database,
                measurement=measurement,
                field=field,
                filters=InfluxDBProfileManager._create_esdl_filters_from_tags_dict(tags),
                startDate=EDate.from_string(self.start_datetime.strftime('%Y-%m-%dT%H:%M:%S.%f%z')),
                endDate=EDate.from_string(self.end_datetime.strftime('%Y-%m-%dT%H:%M:%S.%f%z')),
            )
            profile_list.append(esdl_profile)

        if len(profile_list) == 1:
            return profile_list[0]
        else:
            return profile_list

    def save_influxdb(self, measurement: str, field_names: list, tags: dict = None):
        """
        Saves profile information to InfluxDB

        :param measurement: name of the InfluxDB measurement where the data must be written to
        :param field_names: list of the fields that should be written to InfluxDB
        :param tags: dictionary with tags and tag values, that should be used when writing this data to InfluxDB
        :return: an esdl.InfluxDBProfile instance or a list of esdl.InfluxDBProfile instances in case multiple fields
                 were specified, with proper references to the data in the database
        """
        json_body = []
        if not measurement or measurement == "":
            raise NoDataException("Measurement name is not specified, so no data can be written to InfluxDB")
        if not field_names:
            raise NoDataException("No field names are specified, so no data can be written to InfluxDB")
        if tags and not isinstance(tags, dict):
            raise WrongTagsFormatException("The tags parameter should be a dictionary with tag names as keys")
        for profile_row in self.profile_data_list:
            dt_string = profile_row[0].strftime('%Y-%m-%dT%H:%M:%S%z')
            fields = {}
            for fld_name in field_names:
                if fld_name in self.profile_header:
                    idx = self.profile_header.index(fld_name)
                    fields[self.profile_header[idx]] = profile_row[idx]
                else:
                    raise UnknownProfileNameException(
                        f"Given field name ({fld_name}) is not recognized as a profile name"
                    )

            if fields:
                data = {
                    "measurement": measurement,
                    "time": dt_string,
                    "fields": fields
                }
                if tags:
                    data["tags"] = tags

                json_body.append(data)

        if json_body:
            self.influxdb_client.write_points(points=json_body, database=self.database_settings.database, batch_size=100)
            return self.get_esdl_influxdb_profile(measurement, field_names, tags)

        else:
            return None


class NoDataException(Exception):
    """Thrown when no profile data can be written to InfluxDB"""
    pass


class WrongFilterFormatException(Exception):
    """Thrown when filter specification in ESDL profile cannot be parsed"""
    pass


class WrongTagsFormatException(Exception):
    """Thrown when the tags parameter has the wrong structure"""
    pass



#  This work is based on original code developed and copyrighted by TNO 2023.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO
from typing import Dict, Any

from pyecore.resources import ResourceSet, URI
from pyecore.ecore import EEnum, EAttribute, EOrderedSet, EObject
from pyecore.utils import alias
from pyecore.resources.resource import HttpURI, Resource
from esdl.resources.xmlresource import XMLResource
from esdl import esdl
from uuid import uuid4
from io import BytesIO
from esdl import support_functions


class EnergySystemHandler:

    def __init__(self, energy_system=None):
        if energy_system is not None:
            self.energy_system = energy_system
        self.resource = None
        self.rset = ResourceSet()

        # Assign files with the .esdl extension to the XMLResource instead of default XMI
        self.rset.resource_factory['esdl'] = lambda uri: XMLResource(uri)

        # fix python builtin 'from' that is also used in ProfileElement as attribute
        # use 'start' instead of 'from' when using a ProfileElement
        # alias('start', esdl.ProfileElement.findEStructuralFeature('from'))
        esdl.ProfileElement.from_.name = 'from'
        setattr(esdl.ProfileElement, 'from', esdl.ProfileElement.from_)
        alias('start', esdl.ProfileElement.from_)

        setattr(EObject, '__copy__', support_functions.clone)
        setattr(EObject, 'clone', support_functions.clone)
        setattr(EObject, '__deepcopy__', support_functions.deepcopy)
        setattr(EObject, 'deepcopy', support_functions.deepcopy)

        # have a nice __repr__ for the EnergySystem class when printing. As most objects are linked together
        # we don't do this for other classes, due to recursion.
        esdl.EnergySystem.__repr__ = lambda x: '{}({})'.format(type(x).__name__, EnergySystemHandler.attr_to_dict(x))

    def _new_resource_set(self):
        """Resets the resourceset and removes all files that are loaded before (e.g. when loading a new file)"""
        self.rset = ResourceSet()
        self.resource = None
        # Assign files with the .esdl extension to the XMLResource instead of default XMI
        self.rset.resource_factory['esdl'] = lambda uri: XMLResource(uri)

    def load_file(self, uri_or_filename: str) -> esdl.EnergySystem:
        """Loads a file in a new resource set"""
        if uri_or_filename[:4] == 'http':
            uri = HttpURI(uri_or_filename)
        else:
            uri = URI(uri_or_filename)
        return self.load_uri(uri)

    def load_uri(self, uri) -> esdl.EnergySystem:
        """Loads a new resource in a new resourceSet"""
        self._new_resource_set()
        self.resource = self.rset.get_resource(uri)
        # At this point, the model instance is loaded!
        self.energy_system = self.resource.contents[0]
        return self.energy_system

    def add_uri(self, uri) -> esdl.EnergySystem:
        """Adds the specified URI to the resource set, i.e. load extra resources that the resource can refer to."""
        self.resource = self.rset.get_resource(uri)
        # At this point, the model instance is loaded!
        self.energy_system = self.resource.contents[0]
        return self.energy_system

    def load_from_string(self, esdl_string) -> esdl.EnergySystem:
        uri = StringURI('from_string.esdl', esdl_string)
        # self._new_resource_set()
        self.resource = self.rset.create_resource(uri)
        self.resource.load()
        self.energy_system = self.resource.contents[0]
        return self.energy_system

    def to_string(self) -> str:
        """
        To serialize an energy system in a resource, we use a StringURI to save it to a string instead of a file
        :return: XML string of the current resource/ESDL file that is loaded.
        """
        uri = StringURI('to_string.esdl')
        self.resource.save(uri)
        # return the string
        return uri.getvalue()

    def to_bytesio(self) -> BytesIO:
        """Returns a BytesIO stream for the energy system"""
        uri = StringURI('to_string.esdl')
        self.resource.save(uri)
        return uri.get_stream()

    def get_external_reference(self, url: str, object_id: str = None):
        """
        Support function to create a reference to an external ESDL/EDD file.

        :param url: web address e.g. a URL from the EDR, https://drive.esdl.hesi.energy/store/resource/edr/Public/Key figures/Emissiefactoren energiedragers 2017.edd
        :param object_id: Optional, the id of the object in the file to refer to, if not specified, it will return the root of the file.
        :return: EObject returning the root of the document, or the object that is specified by the id found in that file

        Be aware that when assigning an external reference to a containment relation, the content from the external reference
        will be moved (not copied) to the resource that it is assigned to. If you want the reference to be external,
        only use this for references (no diamond), not containment relations (a diamond in the ESDL UML diagram)
        """
        remote_resource = self.rset.get_resource(url)
        if object_id is not None and object_id in remote_resource.uuid_dict:
            return remote_resource.uuid_dict[object_id]
        else:
            return remote_resource.contents[0]

    def save(self, filename=None):
        """Add the resource to the resourceSet when saving"""
        if filename is None:
            self.resource.save()
        else:
            uri = URI(filename)
            fileresource = self.rset.create_resource(uri)
            # add the current energy system
            fileresource.append(self.energy_system)
            # save the resource
            fileresource.save()
            self.rset.remove_resource(fileresource)

    def save_as(self, filename):
        """Saves the resource under a different filename"""
        self.resource.save(output=filename)

    def get_energy_system(self) -> esdl.EnergySystem:
        return self.energy_system

    # Using this function you can query for objects by ID
    # When loading an ESDL-file, all objects that have an ID defined are stored in resource.uuid_dict automatically
    # Note: If you add things later to the resource, it won't be added automatically to this dictionary though.
    # Use get_by_id_slow() for that
    def get_by_id(self, object_id) -> EObject:
        if object_id in self.resource.uuid_dict:
            return self.resource.uuid_dict[object_id]
        else:
            print('Can\'t find object for id={} in uuid_dict of the ESDL model'.format(object_id))
            raise KeyError('Can\'t find object for id={} in uuid_dict of the ESDL model'.format(object_id))
            return None

    def get_by_id_slow(self, object_id) -> EObject:
        for o in self.energy_system.eAllContents():
            if hasattr(o, 'id'):
                if o.id == object_id:
                    return o

    def update_uuid_dict(self, es: esdl.EnergySystem = None) -> None:
        """
        Update the Resource's uuid_dict of a specific energy system
        This might be necessary when e.g. deepcopying an object in an resource that has references to other parts of that
        resource that are not being deepcopied.
        :param es: energy system that needs a uuid_dict update, if None, the current self.energy_system is used
        """
        if not es:
            es = self.energy_system
        for o in es.eAllContents():
            if hasattr(o, 'id'):
                es.eResource.uuid_dict[o.id] = o

    def add_object(self, obj):
        if hasattr(obj, 'id'):
            if id is not None:
                self.resource.uuid_dict[obj.id] = obj
            else:
                print('Id has not been set for object {}({})', obj.eClass.name, obj)

    def remove_object(self, obj):
        if hasattr(obj, 'id'):
            if id is not None:
                del self.resource.uuid_dict[obj.id]

    def remove_obj_by_id(self, obj_id):
        del self.resource.uuid_dict[obj_id]

    def get_all_instances_of_type(self, esdl_type):
        all_instances = list()
        for esdl_element in self.get_energy_system().eAllContents():
            if isinstance(esdl_element, esdl_type):
                all_instances.append(esdl_element)
        return all_instances

    @staticmethod
    def instantiate_esdltype(className: str) -> EObject:
        """
        Instantiates a new instance of an ESDL class className and returns it.
        E.g. ip:InPort = instantiate_esdltype("InPort")
        """
        return esdl.getEClassifier(className)()

    @staticmethod
    def resolve_fragment(resource: Resource, fragment: str):
        """
        Resolves a URI fragment (e.g. '//@instance.0/@area/@asset.0/@port.0') to the associated object
        and returns the object.
        This is used for objects that have no ID attribute
        """
        return resource.resolve(fragment)

    @staticmethod
    def attr_to_dict(esdl_object) -> Dict[str, Any]:
        """ Creates a dict of all the attributes of an ESDL object, useful for printing/debugging"""
        d = dict()
        d['esdlType'] = esdl_object.eClass.name
        for attr in dir(esdl_object):
            attr_value = esdl_object.eGet(attr)
            if attr_value is not None:
                d[attr] = attr_value
        return d

    @staticmethod
    def generate_uuid() -> str:
        ''' Creates a uuid: useful for generating unique IDs'''
        return str(uuid4())

    def create_empty_energy_system(self, name, es_description=None, inst_title="Instance0", area_title="Main Area") -> esdl.EnergySystem:
        """
        Creates a new empty Energy System (EnergySystem + Instance + Area) and adds it to the resource set
        :param name:
        :param es_description: Optional
        :param inst_title: optional, default "Instance0"
        :param area_title:
        :return: the created empty Energy System
        """
        es_id = self.generate_uuid()
        self.energy_system = esdl.EnergySystem(id=es_id, name=name, description=es_description)

        instance = esdl.Instance(id=self.generate_uuid(), name=inst_title)
        self.energy_system.instance.append(instance)

        # TODO: check if this (adding scope) solves error????
        area = esdl.Area(id=self.generate_uuid(), name=area_title, scope=esdl.AreaScopeEnum.from_string('UNDEFINED'))
        instance.area = area

        self.resource = self.rset.create_resource(StringURI('new_energy_system.esdl'))
        # add the current energy system
        self.resource.append(self.energy_system)
        return self.energy_system

    # Support for Pickling when serializing the energy system in a session
    # The pyEcore classes by default do not allow for simple serialization for Session management in Flask.
    # Internally Flask Sessions use Pickle to serialize a data structure by means of its __dict__. This does not work.
    # Furthermore, ESDL can contain cyclic relations. Therefore we serialize to XMI and back if necessary.
    # This is not efficient with large ESDLs
    def __getstate__(self):
        state = dict()
        state['energySystem'] = self.to_string();
        return state

    def __setstate__(self, state):
        self.__init__()
        self.load_from_string(state['energySystem'])

    @staticmethod
    def version():
        '''Returns the version of pyESDL'''
        return __version__


'''Version information about pyESDL'''
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


"""
Allows to convert a model to an XML string and vice versa
"""
class StringURI(URI):
    def __init__(self, uri, text=None):
        """
        Creates a new StringURI to be used for a resource. By setting text it can convert that XML text to a
        python class model, when doing a resource.load(uri). If text is empty and the associated resource is not,
        resource.save(uri) will convert the resource to XML. Use uri.getvalue() to return the string representation of
        this resource.
        :param uri: the file name of this ESDL, should be unique in this resource set.
        :param text: XML string to convert to a resource (pyEcore classes)
        """
        super(StringURI, self).__init__(uri)
        if text is not None:
            self.__stream = BytesIO(text.encode('UTF-8'))

    def getvalue(self):
        """
        :return: the XML as a string of this ESDL resource
        """
        readbytes = self.__stream.getvalue()
        # somehow stringIO does not work, so we use BytesIO
        string = readbytes.decode('UTF-8')
        return string

    def create_instream(self):
        return self.__stream

    def create_outstream(self):
        self.__stream = BytesIO()
        return self.__stream

    def get_stream(self):
        return self.__stream



#  This work is based on original code developed and copyrighted by TNO 2023.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO

from typing import Union, Type

from pyecore.ecore import EObject, EClass, MetaEClass

from esdl import esdl
from esdl.ecore_documentation import EcoreDocumentation

__esdl_ecore_documentation = None


"""
Convert between ESDL quantities and units, e.g. MW to kW or EUR/kW to MEUR/GW 
including some convertable units, e.g. Joule to Wh and Kelvin to Celsius)
"""

POWER_IN_TW = esdl.QuantityAndUnitType(description="Power in TW", id="POWER_in_TW",
                                       physicalQuantity=esdl.PhysicalQuantityEnum.POWER,
                                       unit=esdl.UnitEnum.WATT,
                                       multiplier=esdl.MultiplierEnum.TERA)
"""Power in TW [QuantityAndUnitType]"""

POWER_IN_GW = esdl.QuantityAndUnitType(description="Power in GW", id="POWER_in_GW",
                                       physicalQuantity=esdl.PhysicalQuantityEnum.POWER,
                                       unit=esdl.UnitEnum.WATT,
                                       multiplier=esdl.MultiplierEnum.GIGA)
"""Power in GW [QuantityAndUnitType]"""

POWER_IN_MW = esdl.QuantityAndUnitType(description="Power in MW", id="POWER_in_MW",
                                       physicalQuantity=esdl.PhysicalQuantityEnum.POWER,
                                       unit=esdl.UnitEnum.WATT,
                                       multiplier=esdl.MultiplierEnum.MEGA)
"""Power in MW [QuantityAndUnitType]"""

POWER_IN_kW = esdl.QuantityAndUnitType(description="Power in kW", id="POWER_in_kW",
                                      physicalQuantity=esdl.PhysicalQuantityEnum.POWER,
                                      unit=esdl.UnitEnum.WATT,
                                      multiplier=esdl.MultiplierEnum.KILO)
"""Power in kW [QuantityAndUnitType]"""

POWER_IN_W = esdl.QuantityAndUnitType(description="Power in Watt", id="POWER_in_W",
                                      physicalQuantity=esdl.PhysicalQuantityEnum.POWER,
                                      unit=esdl.UnitEnum.WATT)
"""Power in W [QuantityAndUnitType]"""

ENERGY_IN_PJ = esdl.QuantityAndUnitType(description="Energy in PJ", id="ENERGY_in_PJ",
                                        physicalQuantity=esdl.PhysicalQuantityEnum.ENERGY,
                                        unit=esdl.UnitEnum.JOULE,
                                        multiplier=esdl.MultiplierEnum.PETA)
"""Energy in PJ [QuantityAndUnitType]"""

ENERGY_IN_TJ = esdl.QuantityAndUnitType(description="Energy in TJ", id="ENERGY_in_TJ",
                                        physicalQuantity=esdl.PhysicalQuantityEnum.ENERGY,
                                        unit=esdl.UnitEnum.JOULE,
                                        multiplier=esdl.MultiplierEnum.TERA)
"""Energy in TJ [QuantityAndUnitType]"""

ENERGY_IN_GJ = esdl.QuantityAndUnitType(description="Energy in GJ", id="ENERGY_in_GJ",
                                        physicalQuantity=esdl.PhysicalQuantityEnum.ENERGY,
                                        unit=esdl.UnitEnum.JOULE,
                                        multiplier=esdl.MultiplierEnum.GIGA)
"""Energy in GJ [QuantityAndUnitType]"""

ENERGY_IN_MJ = esdl.QuantityAndUnitType(description="Energy in MJ", id="ENERGY_in_MJ",
                                        physicalQuantity=esdl.PhysicalQuantityEnum.ENERGY,
                                        unit=esdl.UnitEnum.JOULE,
                                        multiplier=esdl.MultiplierEnum.MEGA)
"""Energy in MJ [QuantityAndUnitType]"""

ENERGY_IN_kJ = esdl.QuantityAndUnitType(description="Energy in kJ", id="ENERGY_in_kJ",
                                        physicalQuantity=esdl.PhysicalQuantityEnum.ENERGY,
                                        unit=esdl.UnitEnum.JOULE,
                                        multiplier=esdl.MultiplierEnum.KILO)
"""Energy in kJ [QuantityAndUnitType]"""

ENERGY_IN_J = esdl.QuantityAndUnitType(description="Energy in J", id="ENERGY_in_J",
                                        physicalQuantity=esdl.PhysicalQuantityEnum.ENERGY,
                                        unit=esdl.UnitEnum.JOULE,
                                        multiplier=esdl.MultiplierEnum.NONE)
"""Energy in J [QuantityAndUnitType]"""

ENERGY_IN_TWh = esdl.QuantityAndUnitType(description="Energy in TWh", id="ENERGY_in_TWh",
                                         physicalQuantity=esdl.PhysicalQuantityEnum.ENERGY,
                                         unit=esdl.UnitEnum.WATTHOUR,
                                         multiplier=esdl.MultiplierEnum.TERA)
"""Energy in TWh [QuantityAndUnitType]"""

ENERGY_IN_GWh = esdl.QuantityAndUnitType(description="Energy in GWh", id="ENERGY_in_GWh",
                                         physicalQuantity=esdl.PhysicalQuantityEnum.ENERGY,
                                         unit=esdl.UnitEnum.WATTHOUR,
                                         multiplier=esdl.MultiplierEnum.GIGA)
"""Energy in GWh [QuantityAndUnitType]"""

ENERGY_IN_MWh = esdl.QuantityAndUnitType(description="Energy in MWh", id="ENERGY_in_MWh",
                                         physicalQuantity=esdl.PhysicalQuantityEnum.ENERGY,
                                         unit=esdl.UnitEnum.WATTHOUR,
                                         multiplier=esdl.MultiplierEnum.MEGA)
"""Energy in MWh [QuantityAndUnitType]"""

ENERGY_IN_kWh = esdl.QuantityAndUnitType(description="Energy in kWh", id="ENERGY_in_kWh",
                                         physicalQuantity=esdl.PhysicalQuantityEnum.ENERGY,
                                         unit=esdl.UnitEnum.WATTHOUR,
                                         multiplier=esdl.MultiplierEnum.KILO)
"""Energy in kWh [QuantityAndUnitType]"""

ENERGY_IN_Wh = esdl.QuantityAndUnitType(description="Energy in Wh", id="ENERGY_in_Wh",
                                         physicalQuantity=esdl.PhysicalQuantityEnum.ENERGY,
                                         unit=esdl.UnitEnum.WATTHOUR,
                                         multiplier=esdl.MultiplierEnum.NONE)
"""Energy in Wh [QuantityAndUnitType]"""

COST_IN_MEur = esdl.QuantityAndUnitType(description="Cost in MEur", id="COST_in_MEUR",
                                        physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                                        unit=esdl.UnitEnum.EURO,
                                        multiplier=esdl.MultiplierEnum.MEGA)
"""Cost in Million Euro [QuantityAndUnitType]"""

COST_IN_Eur_per_MWh = esdl.QuantityAndUnitType(description="Cost in €/MWh", id="COST_in_EURperMWH",
                                        physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                                        unit=esdl.UnitEnum.EURO,
                                        perMultiplier=esdl.MultiplierEnum.MEGA,
                                        perUnit=esdl.UnitEnum.WATTHOUR)
"""Cost in Euro per MWh [QuantityAndUnitType]"""

COST_IN_Eur_per_GJ = esdl.QuantityAndUnitType(description="Cost in €/GJ", id="COST_in_EURperGJ",
                                        physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                                        unit=esdl.UnitEnum.EURO,
                                        perMultiplier=esdl.MultiplierEnum.GIGA,
                                        perUnit=esdl.UnitEnum.JOULE)
"""Cost in Euro per GigaJoule [QuantityAndUnitType]"""

COST_IN_MEur_per_GW_per_year = esdl.QuantityAndUnitType(description="Cost in M€/GW/yr", id="COST_in_MEURperGWperYear",
                                        physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                                        multiplier=esdl.MultiplierEnum.MEGA,
                                        unit=esdl.UnitEnum.EURO,
                                        perMultiplier=esdl.MultiplierEnum.GIGA,
                                        perUnit=esdl.UnitEnum.WATT,
                                        perTimeUnit=esdl.TimeUnitEnum.YEAR)
"""Operational costs (OPEX) in MEUR/GW/yr [QuantityAndUnitType]"""

COST_IN_MEur_per_GW = esdl.QuantityAndUnitType(description="Cost in M€/GW", id="COST_in_MEURperGW",
                                        physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                                        multiplier=esdl.MultiplierEnum.MEGA,
                                        unit=esdl.UnitEnum.EURO,
                                        perMultiplier=esdl.MultiplierEnum.GIGA,
                                        perUnit=esdl.UnitEnum.WATT)
"""Installation cost (CAPEX) in MEUR/GW [QuantityAndUnitType]"""

COST_IN_MEur_per_PJ = esdl.QuantityAndUnitType(description="Cost in M€/PJ", id="COST_in_MEURperPJ",
                                        physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                                        multiplier=esdl.MultiplierEnum.MEGA,
                                        unit=esdl.UnitEnum.EURO,
                                        perMultiplier=esdl.MultiplierEnum.GIGA,
                                        perUnit=esdl.UnitEnum.JOULE)
"""Variable cost in MEUR/PJ [QuantityAndUnitType]"""


def equals(base_unit: esdl.QuantityAndUnitType, other: esdl.QuantityAndUnitType) -> bool:
    """Checks if two units are equal based on physical quantity, multiplier, perMultiplier and perUnit attributes"""
    if base_unit.unit == other.unit and \
            base_unit.multiplier == other.multiplier and \
            base_unit.perUnit == other.per_unit and \
            base_unit.perMultiplier == other.perMultiplier and \
            base_unit.physicalQuantity == other.physicalQuantity:
        return True
    return False


def convertable(source: esdl.UnitEnum, target: esdl.UnitEnum) -> bool:
    """Checks if a unit can be converted to another unit, e.g. Joule -> Wh or Kelvin -> Celsius"""
    try:
        convert_unit(0, source, target)  # check if convert_unit does not raise an exception
        return True
    except UnitException as e:
        return False


def same_physical_quantity(source: esdl.QuantityAndUnitType, target: esdl.QuantityAndUnitType) -> bool:
    """Returns if two esdl.QuantityAndUnitType are equal or if the unit is convertable (e.g. J to Wh)"""
    return source.physicalQuantity == target.physicalQuantity \
        and source.perTimeUnit == target.perTimeUnit \
        and (source.unit == target.unit and source.perUnit == target.perUnit) or \
            (convertable(source.unit, target.unit) and convertable(source.perUnit, target.perUnit))


def convert_to_unit(value: float, source_unit: esdl.AbstractQuantityAndUnit, target_unit: esdl.AbstractQuantityAndUnit) -> float:
    """
    Converts a value from the source_unit into a target_unit

    example: converted_value: float = convert_to_unit(10, ENERGY_IN_J, ENERGY_IN_MWh)

    :param value: the value that needs to be converted
    :param source_unit: the QuantityAndUnit of the source (can also be a reference)
    :param target_unit: the QuantityAndUnit of the target (can also be a reference, they are resolved to a Type)
    :return: the converted value
    """
    if source_unit is None or target_unit is None:
        raise UnitException(f'Missing source unit in unit conversion: source:{source_unit}, target:{target_unit}')
    while isinstance(source_unit, esdl.QuantityAndUnitReference):  # resolve QaU references if necessary
        source_unit = source_unit.reference
    while isinstance(target_unit, esdl.QuantityAndUnitReference):  # resolve QaU references if necessary
        target_unit = target_unit.reference
    if same_physical_quantity(target_unit, source_unit):
        return convert_unit(
            convert_unit(
                convert_multiplier(source_unit, target_unit) * value, source_unit.unit, target_unit.unit), source_unit.perUnit, target_unit.perUnit)

    else:
        raise UnitException(f'Conversion mismatch between units: source_unit={source_unit}, target_unit={target_unit}')


def convert_multiplier(source: esdl.QuantityAndUnitType, target: esdl.QuantityAndUnitType) -> float:
    """
    Calculates the factor between the source unit and the target unit
    """
    value = multipier_value(source.multiplier) / multipier_value(target.multiplier) * \
        multipier_value(target.perMultiplier) / multipier_value(source.perMultiplier)
    #print(f"{multipier_value(source.multiplier)} / {multipier_value(target.multiplier)} * {multipier_value(target.perMultiplier)} / {multipier_value(source.perMultiplier)}")
    #print(f"Converting source {source} to {target}: factor={value}")
    return value


    # MultiplierEnum
    # ['NONE', 'ATTO', 'FEMTO', 'PICO', 'NANO', 'MICRO',
    #  'MILLI', 'CENTI', 'DECI', 'DEKA', 'HECTO', 'KILO', 'MEGA',
    #  'GIGA', 'TERA', 'TERRA', 'PETA', 'EXA']
factors = [1, 1E-18, 1E-15, 1E-12, 1E-9, 1E-6, 1E-3, 1E-2, 1E-1, 1E1,
               1E2, 1E3, 1E6, 1E9, 1E12, 1E12, 1E15, 1E18]


def multipier_value(multiplier: esdl.MultiplierEnum):
    """Converts ESDL MultiplierEnum into a numeric value."""
    return factors[esdl.MultiplierEnum.eLiterals.index(multiplier)]


unit_mapping = {
    esdl.UnitEnum.WATTHOUR: {esdl.UnitEnum.JOULE: {'type': 'MULTIPLY', 'value': 3600.0}},
    esdl.UnitEnum.JOULE: {esdl.UnitEnum.WATTHOUR: {'type': 'MULTIPLY', 'value': 1.0/3600.0}},
    esdl.UnitEnum.DEGREES_CELSIUS: {esdl.UnitEnum.KELVIN: {'type': 'ADDITION', 'value': 273.15}},
    esdl.UnitEnum.KELVIN: {esdl.UnitEnum.DEGREES_CELSIUS: {'type': 'ADDITION', 'value': -273.15}},
    esdl.UnitEnum.TONNE: {esdl.UnitEnum.GRAM: {'type': 'MULTIPLY', 'value': 1E6}},
    esdl.UnitEnum.GRAM: {esdl.UnitEnum.TONNE: {'type': 'MULTIPLY', 'value': 1E-6}},
}


def convert_unit(value: float, source_quantity_unit: esdl.UnitEnum, target_quantity_unit: esdl.UnitEnum) -> float:
    """
    Does some basic unit conversion, only Joule to Wh, Wh to Joule and \N{DEGREE SIGN}C to Kelvin and vice versa.
    Can only convert units when physical quantities are the same (e.g. Energy, Temperature)
    """
    if source_quantity_unit == target_quantity_unit:
        return value
    else:
        if source_quantity_unit in unit_mapping:
            source_map = unit_mapping[source_quantity_unit]
            if target_quantity_unit in source_map:
                conversion = source_map[target_quantity_unit]
                if conversion['type'] == 'MULTIPLY':
                    #print(f"Unit conversion factor {source_quantity_unit.name} to {target_quantity_unit.name} factor={conversion['value']}")
                    return value * conversion['value']
                elif conversion['type'] == 'ADDITION':
                    return value + conversion['value']
            else:
                raise UnitException(f"No mapping available from {source_quantity_unit.name} to {target_quantity_unit.name}")
        else:
            raise UnitException(f"Cannot convert {source_quantity_unit.name} into {target_quantity_unit.name}")


class UnitException(Exception):
    """Thrown when two esdl.QuantityAndUnitTypes can not be converted"""
    pass


def get_attribute_unit(esdl_class: Union[EClass, EObject, Type, str], attribute: str) -> str:
    """
    Retrieves the unit defined in the ESDL ecore schema for a specific attribute

    example: get_attribute_unit("PowerPlant", "power"), or
             get_attribute_unit(esdl.PowerPlant, "power"), or
             get_attribute_unit(myPowerPlant, "power")

    :param esdl_class: the ESDL class (as string, ESDL class or ESDL Object)
    :param attribute: string describing the attribute of the esdl_class
    :return: a string with the unit (e.g. W, m, years)
    """
    global __esdl_ecore_documentation
    if __esdl_ecore_documentation is None:
        __esdl_ecore_documentation = EcoreDocumentation()
    esdl_class_str = esdl_class
    if isinstance(esdl_class, EObject):
        # noinspection PyUnresolvedReferences
        esdl_class_str = esdl_class.eClass.name
    elif isinstance(esdl_class, MetaEClass):
        # noinspection PyUnresolvedReferences
        esdl_class_str = esdl_class.eClass.name
    return __esdl_ecore_documentation.get_unit(className=esdl_class_str, attributeName=attribute)


#  This work is based on original code developed and copyrighted by TNO 2023.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO
import locale

import openpyxl
import pytz

from esdl.profiles.profilemanager import ProfileManager, DuplicateValueInProfileException, ProfileType
from openpyxl import Workbook

from esdl.utils.datetime_utils import parse_date


class ExcelProfileManager(ProfileManager):
    """
    ExcelProfileManager: manages profile data that can be loaded from and saved to excel.

    ExcelProfileManager is a subclass of ProfileManager, so it also provides functionality to convert from/to
    different ESDL profiles and to load/save to CSV
    """
    def __init__(self, source_profile=None):
        """
        Constructor of the ExcelProfileManager

        :param source_profile: the source profile of which the data is copied into this profiles manager instance
        """
        super().__init__()

        if source_profile:
            self.convert(source_profile)

    def load_excel(self, file_path, sheet_name: str = None):
        """
        Reads profile data from an Excel file.

        :param file_path: the path to the Excel file
        :param sheet_name: the name of the sheet that contains the data. By default, the first sheet is used
        """
        self.clear_profile()
        self.profile_type = ProfileType.EXCEL

        wb = openpyxl.load_workbook(file_path)
        if sheet_name:
            wb.active = sheet_name
        sheet = wb.active

        self.profile_header = [cell.value for cell in sheet[1]]
        self.num_profile_items = 0

        previous_datetime = None
        for row in sheet.iter_rows(min_row=2):
            try:
                dt = parse_date(row[0].value)
            except ValueError:  # ValueError: row in Excel doesn't start with recognizable datetime value, ignore row
                continue

            try:
                aware_dt = pytz.utc.localize(dt)  # Assume timezone is UTC if no TZ was given
            except ValueError:  # ValueError: No naive datetime (tzinfo is already set)
                aware_dt = dt
            row[0].value = aware_dt
            for elem_idx in range(1, len(row)):
                try:
                    row[elem_idx].value = float(row[elem_idx].value)      # Assume float values
                except Exception as e:
                    raise Exception(f"Cannot parse profile value in CSV ({row[elem_idx].value})")

            if previous_datetime:
                if previous_datetime == aware_dt:
                    raise DuplicateValueInProfileException(
                        "CSV contains duplicate datetimes ({}). Check timezone and daylight saving".
                        format(aware_dt.strftime('%Y-%m-%dT%H:%M:%S%z'))
                    )
            previous_datetime = aware_dt

            if self.start_datetime is None:
                self.start_datetime = aware_dt

            self.profile_data_list.append([cell.value for cell in row])
            self.num_profile_items += 1

        self.end_datetime = self.profile_data_list[-1][0]

    def save_excel(self, file_path, sheet_name: str = "Sheet1"):
        """
        Saves profile data to an Excel file.

        :param file_path: the path to the Excel file
        :param sheet_name: the name of the sheet that should be created to store the data (default is "Sheet1")
        """

        wb = Workbook()
        ws = wb.active
        ws.title = sheet_name

        ws.append(self.profile_header)
        for row in self.profile_data_list:
            out_row = row[:]
            out_row[0] = out_row[0].strftime('%Y-%m-%dT%H:%M:%SZ')
            ws.append(out_row)

        wb.save(file_path)



#  This work is based on original code developed and copyrighted by TNO 2023.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO

from pyecore.resources import ResourceSet, URI
from pyecore.utils import DynamicEPackage
from pyecore.ecore import EModelElement, EAnnotation
from pyecore.resources.resource import HttpURI

import esdl
from esdl.resources.xmlresource import XMLResource


esdlVersion = "unknown"

class EcoreDocumentation:
    """This class loads the dynamic meta-model and returns the documentation for attributes as these are
    not present in the static meta-model (e.g. in the classes that are generated in esdl.py)"""
    def __init__(self, esdlEcoreFile=None):
            self.esdl = None
            self.rset = None
            self.esdl_model = None
            self.resource = None
            if esdlEcoreFile is None:
                self.esdlEcoreFile = 'https://raw.githubusercontent.com/EnergyTransition/ESDL/master/esdl/model/esdl.ecore'
            else:
                self.esdlEcoreFile = esdlEcoreFile
            self._init_metamodel()
            global esdlVersion
            esdlVersion = self.get_esdl_version()

    def _init_metamodel(self):
            self.rset = ResourceSet()

            # Assign files with the .esdl extension to the XMLResource instead of default XMI
            self.rset.resource_factory['esdl'] = lambda uri: XMLResource(uri)

            # Read esdl.ecore as meta model
            mm_uri = URI(self.esdlEcoreFile)
            if self.esdlEcoreFile[:4] == 'http':
                mm_uri = HttpURI(self.esdlEcoreFile)
            try:
                esdl_model_resource = self.rset.get_resource(mm_uri)
            except Exception as e:
                self.esdl = esdl.esdl # use imported esdl packages (that miss attribute documentation)
                self.esdl_model = esdl.eClass
                return

            esdl_model = esdl_model_resource.contents[0]
            self.esdl_model = esdl_model
            # logger.debug('Namespace: {}'.format(esdl_model.nsURI))
            self.rset.metamodel_registry[esdl_model.nsURI] = esdl_model

            # Create a dynamic model from the loaded esdl.ecore model, which we can use to build Energy Systems
            self.esdl = DynamicEPackage(esdl_model)


    def get_doc(self, className, attributeName):
        """ Returns the documentation of an attribute from the dynamic meta model,
        because the static meta model does not contain attribute documentation"""
        ecoreClass = self.esdl_model.getEClassifier(className)
        if ecoreClass is None: return None
        attr = ecoreClass.findEStructuralFeature(attributeName)
        if attr is None: return None
        # logger.debug('Retrieving doc for {}: {}'.format(attributeName, attr.__doc__))
        return attr.__doc__

    def get_unit(self, className:str, attributeName:str) -> str:
        """
        :value: EmodelElement (e.g. EObject) that has an annotation attached
        :returns: The unit attached to this annotation or empty string
        """
        ecoreClass = self.esdl_model.getEClassifier(className)
        if ecoreClass is None: return None
        attr = ecoreClass.findEStructuralFeature(attributeName)
        if attr is None: return None

        annotation: EAnnotation = attr.getEAnnotation('http://www.tno.nl/esdl/attribute/unit')
        unit = annotation.details.get('unit', '') if annotation else None
        return unit if unit else ''

    def get_esdl_version(self):
        annotation: EAnnotation = self.esdl_model.getEAnnotation('http://www.tno.nl/esdl/version')
        version = annotation.details.get('version', '') if annotation else None
        return version if version else 'unknown'



#  This work is based on original code developed and copyrighted by TNO 2023.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO

from datetime import datetime
from contextlib import suppress


def parse_date(str_date: str):
    """
    Function to parse a date or datetime string to a datetime object. Tries many different formats to see
    if it can make sense out of the string. Both dates and times seperated by a 'T' and a space are supported.
    Both datetimes with and without timezone specification are supported

    :param str_date: input string representing a date or a datetime
    :return: datetime object based on the information in the input string
    """
    try:
        return datetime.fromisoformat(str_date)
    except Exception:
        formats = ('%Y-%m-%dT%H:%M:%S.%f%z',
                   '%Y-%m-%dT%H:%M:%S.%f%Z',
                   '%Y-%m-%dT%H:%M:%S.%f',
                   '%Y-%m-%dT%H:%M:%S%z',
                   '%Y-%m-%dT%H:%M:%S%Z',
                   '%Y-%m-%dT%H:%M:%S',
                   '%Y-%m-%dT%H:%M',
                   '%Y-%m-%d %H:%M:%S.%f%z',
                   '%Y-%m-%d %H:%M:%S.%f%Z',
                   '%Y-%m-%d %H:%M:%S.%f',
                   '%Y-%m-%d %H:%M:%S%z',
                   '%Y-%m-%d %H:%M:%S%Z',
                   '%Y-%m-%d %H:%M:%S',
                   '%Y-%m-%d %H:%M',
                   '%Y-%m-%d',

                   # As a last resort, also try Dutch formatting of dates DD-MM-YYYY
                   '%d-%m-%YT%H:%M:%S.%f%z',
                   '%d-%m-%YT%H:%M:%S.%f%Z',
                   '%d-%m-%YT%H:%M:%S.%f',
                   '%d-%m-%YT%H:%M:%S%z',
                   '%d-%m-%YT%H:%M:%S%Z',
                   '%d-%m-%YT%H:%M:%S',
                   '%d-%m-%YT%H:%M',
                   '%d-%m-%Y %H:%M:%S.%f%z',
                   '%d-%m-%Y %H:%M:%S.%f%Z',
                   '%d-%m-%Y %H:%M:%S.%f',
                   '%d-%m-%Y %H:%M:%S%z',
                   '%d-%m-%Y %H:%M:%S%Z',
                   '%d-%m-%Y %H:%M:%S',
                   '%d-%m-%Y %H:%M',
                   '%d-%m-%Y',
                   )
        for fmt in formats:
            with suppress(ValueError):
                return datetime.strptime(str_date, fmt)
        raise ValueError('Date format is unknown')



"""
The json module introduces JSON resource and JSON parsing.
This is the json module from pyEcore 0.15
"""
from enum import unique, Enum
from functools import lru_cache
import json
from pyecore.resources.resource import Resource
from pyecore.ecore import EObject, EProxy, ECollection, EClass, EEnumLiteral, EStringToStringMapEntry


@unique
class JsonOptions(Enum):
    SERIALIZE_DEFAULT_VALUES = 0
    ENCODER = 1
    DECODER = 2


NO_OBJECT = object()


class JsonResource(Resource):
    def __init__(self, uri=None, use_uuid=False, indent=None, ref_tag='$ref'):
        super().__init__(uri, use_uuid)
        self._resolve_later = []
        self._load_href = {}
        self.indent = indent
        self.ref_tag = ref_tag
        self.mappers = {}
        self.default_mapper = DefaultObjectMapper()

    def load(self, options=None):
        self.options = options or {}
        self.cache_enabled = True
        json_value = self.uri.create_instream()

        decoder = self.options.get(JsonOptions.DECODER)
        d = json.loads(json_value.read().decode('utf-8'), cls=decoder)

        if isinstance(d, list):
            for x in d:
                self.to_obj(x, first=True)
        else:
            self.to_obj(d, first=True)
        self.uri.close_stream()
        for inst, refs in self._load_href.items():
            self.process_inst(inst, refs)
        self._load_href.clear()
        self._find_feature.cache_clear()
        self.cache_enabled = False

    def save(self, output=None, options=None):
        self.options = options or {}
        stream = self.open_out_stream(output)
        dict_list = []
        for root in self.contents:
            dict_list.append(self.to_dict(root))
        if len(dict_list) <= 1:
            dict_list = dict_list[0]

        encoder = self.options.get(JsonOptions.ENCODER)
        stream.write(json.dumps(dict_list, indent=self.indent, cls=encoder)
                     .encode('utf-8'))

        stream.flush()
        self.uri.close_stream()
        self.options = None

    def _uri_fragment(self, obj):
        if obj.eResource == self:
            use_id = self.use_uuid
        else:
            use_id = obj.eResource and obj.eResource.use_uuid
        if use_id:
            self._assign_uuid(obj)
            return obj._internal_id
        else:
            return obj.eURIFragment()

    @staticmethod
    def serialize_eclass(eclass):
        return f'{eclass.eRoot().nsURI}{eclass.eURIFragment()}'

    def register_mapper(self, eclass, mapper_class):
        if hasattr(eclass, 'python_class'):
            eclass = eclass.python_class
        self.mappers[eclass] = mapper_class

    def object_uri(self, obj):
        # if obj.eResource == self:
        #     resource_uri = ''
        # else:
        #     resource_uri = self.eResource.uri if obj.eResource else ''
        return self._build_path_from(obj)[0]

    def _to_ref_from_obj(self, obj, opts=None, use_uuid=None, resource=None, **kwargs):
        uri = self.serialize_eclass(obj.eClass)
        ref = {'eClass': uri}
        ref[self.ref_tag] = self.object_uri(obj)
        return ref

    # @singledispatchmethod
    # def to_dict(self, obj, is_ref=False):
    #     return obj
    #
    # @to_dict.register(type)
    # def _(self, obj, is_ref=False):
    #     if not issubclass(obj, EObject):
    #         return obj
    #     if is_ref:
    #         fun = self._to_ref_from_obj
    #         return fun(obj.eClass, self.options, self.use_uuid, self)
    #
    # @to_dict.register(EEnumLiteral)
    # def _(self, obj, is_ref=False):
    #     return obj.name
    #
    # @to_dict.register(EObject)
    # def _(self, obj, is_ref=False):
    #     if is_ref:
    #         fun = self._to_ref_from_obj
    #     else:
    #         cls = obj.eClass.python_class
    #         mapper = next((self.mappers[k] for k in self.mappers
    #                        if issubclass(cls, k)), self.default_mapper)
    #         fun = mapper.to_dict_from_obj
    #     return fun(obj, self.options, self.use_uuid, self)
    #
    # @to_dict.register(ECollection)
    # def _(self, obj, is_ref=False):
    #     fun = self._to_ref_from_obj if is_ref else self.to_dict
    #     result = []
    #     for x in obj:
    #         write_object = fun(x)
    #         if write_object is NO_OBJECT:
    #             continue
    #         result.append(write_object)
    #     return result

    def to_dict(self, obj, is_noncont_ref=False, is_attr=False, feature=None):
        if isinstance(obj, type) and issubclass(obj, EObject):
            if is_noncont_ref:
                fun = self._to_ref_from_obj
                return fun(obj.eClass, self.options, self.use_uuid, self)
            # else:
            #     cls = obj.python_class
            #     mapper = next((self.mappers[k] for k in self.mappers
            #                    if issubclass(cls, k)), self.default_mapper)
            #     fun = mapper.to_dict_from_obj
        elif isinstance(obj, EEnumLiteral) and is_attr:
            return obj.name
        elif isinstance(obj, EObject):
            if is_noncont_ref:
                fun = self._to_ref_from_obj
            else:
                cls = obj.eClass.python_class
                mapper = next((self.mappers[k] for k in self.mappers
                               if issubclass(cls, k)), self.default_mapper)
                fun = mapper.to_dict_from_obj
            return fun(obj, self.options, self.use_uuid, self)

        elif isinstance(obj, ECollection):
            fun = self._to_ref_from_obj if is_noncont_ref else self.to_dict
            result = []
            for x in obj:
                write_object = fun(x, is_noncont_ref, is_attr, feature=feature)
                if write_object is NO_OBJECT:
                    continue
                result.append(write_object)
            return result
        elif feature._eType.eType in (int, float):
            return obj
        else:
            return feature._eType.to_string(obj)

    @lru_cache()
    def resolve_eclass(self, uri_eclass):
        return self.resolve_object(uri_eclass)

    def to_obj(self, d, owning_feature=None, first=False):
        is_ref = self.ref_tag in d
        if is_ref:
            return EProxy(path=d[self.ref_tag], resource=self)
        excludes = ['eClass', self.ref_tag, 'uuid']
        if 'eClass' in d:
            uri_eclass = d['eClass']
            eclass = self.resolve_eclass(uri_eclass)
        else:
            eclass = owning_feature._eType
        if not eclass:
            raise ValueError(f'Unknown metaclass for uri "{uri_eclass}"')
        if eclass in (EClass.eClass, EClass):
            inst = eclass(d['name'])
            excludes.append('name')
        else:
            inst = eclass()
        if first:
            self.use_uuid = 'uuid' in d
            self.append(inst)

        if self.use_uuid:
            self.uuid_dict[d['uuid']] = inst

        eattributes = []
        containments = []
        ereferences = []
        eclass = inst.eClass
        for key, value in d.items():
            if key in excludes:
                continue
            feature = self._find_feature(eclass, key)
            if not feature:
                raise ValueError(f'Unknown feature {key} for object "{eclass}"')
            if feature.is_attribute:
                eattributes.append((feature, value))
            else:
                if feature.containment:
                    containments.append((feature, value))
                elif owning_feature and feature.eOpposite is not owning_feature:
                    ereferences.append((feature, value))
                elif not feature.eOpposite:
                    ereferences.append((feature, value))
        self.process_inst(inst, eattributes)
        self.process_inst(inst, containments, owning_feature)
        if ereferences:
            self._load_href[inst] = ereferences
        return inst

    def process_inst(self, inst, features, owning_feature=None):
        for feature, value in features:
            if feature._eType is EStringToStringMapEntry and isinstance(value, dict):
                key, val = next(iter(value.items()))
                inst.eGet(feature)[key] = val
            elif isinstance(value, dict):
                element = self.to_obj(value, owning_feature=feature)
                inst.eSet(feature, element)
            elif isinstance(value, list):
                if feature.is_reference:
                    elements = (self.to_obj(x, owning_feature=feature)
                                for x in value)
                    elements = (x for x in elements if x is not None)
                else:
                    elements = (feature._eType.from_string(x) for x in value)
                inst.eGet(feature).extend(list(elements))
            elif isinstance(value, str):
                inst.eSet(feature, feature._eType.from_string(value))
            else:
                # int, float or whatever, we aske the type to parse the value
                inst.eSet(feature, feature._eType.from_string(value))


class DefaultObjectMapper(object):
    def to_dict_from_obj(self, obj, options, use_uuid, resource):
        d = {}
        containingFeature = obj.eContainmentFeature()
        if not containingFeature or obj.eClass is not containingFeature._eType:
            uri = resource.serialize_eclass(obj.eClass)
            d['eClass'] = uri
        for attr in obj._isset:
            if attr.derived or attr.transient:
                continue
            is_ereference = attr.is_reference
            is_ref = is_ereference and not attr.containment
            if is_ereference and attr.eOpposite:
                if attr.eOpposite is containingFeature:
                    continue
            value = obj.eGet(attr)
            serialize_default_option = JsonOptions.SERIALIZE_DEFAULT_VALUES
            if (not options.get(serialize_default_option, False)
                and value == attr.get_default_value()):
                continue
            write_object = resource.to_dict(value,
                                            is_noncont_ref=is_ref,
                                            is_attr=attr.is_attribute,
                                            feature=attr)
            if write_object is not NO_OBJECT:
                d[attr._name] = write_object
            if use_uuid:
                resource._assign_uuid(obj)
                d['uuid'] = obj._internal_id
        return d



#  This work is based on original code developed and copyrighted by TNO 2022.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO
import base64
import json
import urllib
from dataclasses import dataclass

import requests

from esdl.esdl_handler import EnergySystemHandler

EDR_URL="https://drive.esdl.hesi.energy"


@dataclass
class EDRInfo:
    """
    Represents information about an object in the Energy Data Repository (EDR)
    """
    id: str
    """ the ID (or path) of the object in the EDR """
    title: str
    """ the title of the object in the EDR """
    description: str
    """ the description of the object in the EDR """
    esdl_type: str
    """ the ESDL type of the object in the EDR """


class EDRClient:
    """
    Implements an interface to the Energy Data Repository (EDR)
    """
    def __init__(self, host=None, port=None):
        """
        Constructor of the EDR client. Uses the public EDR URL by default, but allows to specify a different "host" and
        "port" for using a local EDR instance

        :param host: alternative hostname of a EDR instance
        :param port: alternative port of a EDR instance
        """
        if host:
            self.EDR_URL = host + ":" + port
        else:
            self.EDR_URL = EDR_URL

    def get_objects_list(self, esdl_type):
        """
        Retrieves a list of objects from the EDR of a given type. The input parameter esdl_type can for example be
        "EnergyAsset", "Sectors", "Carriers" or "InfluxDBProfile". Each element in the returned list is of type EDRInfo
        and has id, title, description and esdl_type fields. The ESDL type of the elements in the list can be subtypes
        of the input esdl_type, e.g. you ask for objects of type "EnergyAsset" and get back objects of type
        "WindTurbine".

        :param esdl_type: the type of the ESDL class
        :return: list of objects of the given type
        """
        url = self.EDR_URL + "/store/edr/query?addESDL=false&addImageData=false&esdlType="+esdl_type+"&maxResults=-1"

        obj_list = list()

        r = requests.get(url)
        if r.status_code == 200:
            result = json.loads(r.text)

            for item in result:
                obj = EDRInfo(
                    id=item["id"],
                    title=item["title"],
                    description=item["description"] if "description" in item else "",
                    esdl_type=item["esdlType"] if "esdlType" in item else None,
                )
                obj_list.append(obj)

        return obj_list

    def get_object_esdl(self, edr_path):
        """
        Retrieves a specific ESDL object from the EDR

        :param edr_path: the path in the EDR to the specific ESDL object
        :return: the ESDL object
        """
        url = self.EDR_URL + "/store/edr/query?addESDL=true&path=" + urllib.parse.quote(edr_path, safe='')

        headers = {
            'Accept': "application/json",
            'User-Agent': "pyESDL/EDRClient"
        }

        r = requests.get(url, headers=headers)
        if r.status_code == 200:
            result = json.loads(r.text)

            if result:
                if 'esdl' in result[0]:
                    esdl_str_b64 = result[0]['esdl']
                    esdl_str_b64_bytes = esdl_str_b64.encode('utf-8')
                    esdl_str_bytes = base64.b64decode(esdl_str_b64_bytes)
                    esdl_str = esdl_str_bytes.decode('utf-8')

                    esh = EnergySystemHandler()
                    esdl_obj = esh.load_from_string(esdl_string=esdl_str)
                    return esdl_obj

        # return None in all other cases
        return None






"""Definition of meta model 'edd'."""
from functools import partial
import pyecore.ecore as Ecore
from pyecore.ecore import *


name = 'edd'
nsURI = 'http://www.tno.nl/edr/edd'
nsPrefix = 'edd'

eClass = EPackage(name=name, nsURI=nsURI, nsPrefix=nsPrefix)

eClassifiers = {}
getEClassifier = partial(Ecore.getEClassifier, searchspace=eClassifiers)


class EnergyDataDescription(EObject, metaclass=MetaEClass):

    id = EAttribute(eType=EString, unique=True, derived=False, changeable=True, iD=True)
    title = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    description = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    tags = EAttribute(eType=EString, unique=True, derived=False, changeable=True, upper=-1)
    version = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    graphURL = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    esdlType = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    lastChanged = EAttribute(eType=EDate, unique=True, derived=False,
                             changeable=True, transient=True)
    publicationDate = EAttribute(eType=EDate, unique=True, derived=False,
                                 changeable=True, transient=True)
    esdl = EReference(ordered=True, unique=True, containment=True, derived=False)
    image = EReference(ordered=True, unique=True, containment=True, derived=False)

    def __init__(self, *, id=None, title=None, description=None, tags=None, esdl=None, image=None, version=None, graphURL=None, esdlType=None, lastChanged=None, publicationDate=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if id is not None:
            self.id = id

        if title is not None:
            self.title = title

        if description is not None:
            self.description = description

        if tags:
            self.tags.extend(tags)

        if version is not None:
            self.version = version

        if graphURL is not None:
            self.graphURL = graphURL

        if esdlType is not None:
            self.esdlType = esdlType

        if lastChanged is not None:
            self.lastChanged = lastChanged

        if publicationDate is not None:
            self.publicationDate = publicationDate

        if esdl is not None:
            self.esdl = esdl

        if image is not None:
            self.image = image


class Image(EObject, metaclass=MetaEClass):

    contentType = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    imageData = EAttribute(eType=EByteArray, unique=True, derived=False, changeable=True)

    def __init__(self, *, contentType=None, imageData=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if contentType is not None:
            self.contentType = contentType

        if imageData is not None:
            self.imageData = imageData



#  This work is based on original code developed and copyrighted by TNO 2022.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO

import geojson
import json
from shapely import wkt, wkb, to_geojson
from shapely.geometry import Point, LineString, Polygon, MultiPolygon, GeometryCollection, shape
from shapely.ops import transform
import esdl
import pyproj


class Shape:
    """
    Represents a shape, like a point, line, polygon or multi polgygon. Provides functionality to convert to and from
    ESDL geometries, Shapely geometries, WKT/WKB, Leaflet (used in the ESDL MapEditor), GeoJSON. Uses a Shapely shape
    internally and provides different methods to convert to and from different other formats.
    """
    def __init__(self):
        """
        Constructor of the Shape class.
        """
        self.shape = None

    @staticmethod
    def create(shape_input):
        """
        Function to create an instance of the right subclass of Shape based on any input. Tries to guess the input
        format and returns an instance of the right class.

        :param shape_input: input that represents information about the shape. Can be a list with coordinates, a
                            dictionary with "lat" and "lng" fields, an ESDL geometry object or a Shapely geometry object.
        :return: an instance of the right Shape subclass
        """
        if isinstance(shape_input, esdl.Point) or (isinstance(shape_input, dict) and "lat" in shape_input):
            return ShapePoint(shape_input)
        if isinstance(shape_input, esdl.Line) or (isinstance(shape_input, list) and "lat" in shape_input[0]):
            return ShapeLine(shape_input)
        if isinstance(shape_input, esdl.Polygon):
            return ShapePolygon(shape_input)
        if isinstance(shape_input, esdl.MultiPolygon):
            return ShapeMultiPolygon(shape_input)

        if isinstance(shape_input, esdl.WKT):
            return Shape.parse_esdl_wkt(shape_input)
        if isinstance(shape_input, esdl.WKB):
            return Shape.parse_esdl_wkb(shape_input)

        if isinstance(shape_input, Point):
            return ShapePoint(shape_input)
        if isinstance(shape_input, LineString):
            return ShapeLine(shape_input)
        if isinstance(shape_input, Polygon):
            return ShapePolygon(shape_input)
        if isinstance(shape_input, MultiPolygon):
            return ShapeMultiPolygon(shape_input)

        if isinstance(shape_input, list) and all(isinstance(elem, list) and "lat" in elem[0] for elem in shape_input):
            return ShapePolygon(shape_input)
        else:
            # TODO: Better check for coordinates structure
            return ShapeMultiPolygon(shape_input)

    @staticmethod
    def parse_esdl(esdl_geometry):
        """
        Function that parses an ESDL geometry object. Must be overridden by the subclass.
        """
        pass

    @staticmethod
    def parse_leaflet(leaflet_coords):
        """
        Function that parses an list of leaflet coordinates. Must be overridden by the subclass.
        """
        pass

    @staticmethod
    def parse_geojson_geometry(geojson_geometry):
        """
        Function that parses a GeoJSON string and creates an instance of the right Shape subclass

        :param geojson_geometry: GeoJSON string
        :return: an instance of the right Shape subclass with the parsed geometry information from the GeoJSON string
        """
        tmp_shp = shape(geojson.loads(json.dumps(geojson_geometry)))
        if isinstance(tmp_shp, Point):
            return ShapePoint(tmp_shp)
        elif isinstance(tmp_shp, LineString):
            return ShapeLine(tmp_shp)
        elif isinstance(tmp_shp, Polygon):
            return ShapePolygon(tmp_shp)
        elif isinstance(tmp_shp, MultiPolygon):
            return ShapeMultiPolygon(tmp_shp)
        else:
            raise Exception("Parsing geojson resulted in unsupported type")

    @staticmethod
    def parse_wkt(wkt_geometry, crs="EPSG:4326"):
        """
        Function that parses a WKT string and creates an instance of the right Shape subclass

        :param wkt_geometry: a WKT (Well Known Text) string
        :return: an instance of the right Shape subclass with the parsed geometry information from the WKT string
        """
        tmp_shp = Shape.transform_crs(wkt.loads(wkt_geometry), crs)

        if isinstance(tmp_shp, Point):
            return ShapePoint(tmp_shp)
        elif isinstance(tmp_shp, LineString):
            return ShapeLine(tmp_shp)
        elif isinstance(tmp_shp, Polygon):
            return ShapePolygon(tmp_shp)
        elif isinstance(tmp_shp, MultiPolygon):
            return ShapeMultiPolygon(tmp_shp)
        elif isinstance(tmp_shp, GeometryCollection):
            return ShapeGeometryCollection(tmp_shp)
        else:
            raise Exception("Parsing WKT resulted in unsupported type")

    @staticmethod
    def parse_wkb(wkb_geometry, crs="EPSG:4326"):
        """
        Function that parses a WKB string and creates an instance of the right Shape subclass

        :param wkb_geometry: a WKB (Well Known Binary) byte string
        :return: an instance of the right Shape subclass with the parsed geometry information from the WKB string
        """
        tmp_shp = Shape.transform_crs(wkb.loads(wkb_geometry), crs)

        if isinstance(tmp_shp, Point):
            return ShapePoint(tmp_shp)
        elif isinstance(tmp_shp, LineString):
            return ShapeLine(tmp_shp)
        elif isinstance(tmp_shp, Polygon):
            return ShapePolygon(tmp_shp)
        elif isinstance(tmp_shp, MultiPolygon):
            return ShapeMultiPolygon(tmp_shp)
        else:
            raise Exception("Parsing WKB resulted in unsupported type")

    @staticmethod
    def parse_esdl_wkt(esdl_wkt):
        """
        Function that parses an esdl.WKT string and creates an instance of the right Shape subclass

        :param esdl_wkt: an instance of esdl.WKT
        :return: an instance of the right Shape subclass with the parsed geometry information from the esdl.WKT
        """
        if isinstance(esdl_wkt, esdl.WKT):
            return Shape.parse_wkt(esdl_wkt.value, esdl_wkt.CRS)
        else:
            raise Exception("Calling parse_esdl_WKT without an esdl.WKT parameter")

    @staticmethod
    def parse_esdl_wkb(esdl_wkb):
        """
        Function that parses an esdl.WKB string and creates an instance of the right Shape subclass

        :param esdl_wkb: an instance of esdl.WKB
        :return: an instance of the right Shape subclass with the parsed geometry information from the esdl.WKB
        """
        if isinstance(esdl_wkb, esdl.WKB):
            return Shape.parse_wkb(esdl_wkb.value, esdl_wkb.CRS)
        else:
            raise Exception("Calling parse_esdl_WKB without an esdl.WKB parameter")

    def get_shape(self):
        """
        Function that returns the Shapely geometry representation of the loaded geometry
        """
        return self.shape

    def get_esdl(self):
        """
        Function that returns an ESDL geometry. Must be overridden by the subclass.
        """
        pass

    def get_wkt(self):
        """
        Function that generates a WKT string from the loaded geometry

        :return: a WKT string representing the loaded geometry information
        """
        return self.shape.wkt

    def get_geojson_feature(self, properties={}):
        """
        Function that generates a GeoJSON feature from the loaded geometry, with the properties given as an input

        :param properties: the properties that must be added to the GeoJSON FFeature
        :return: a GeoJSON Feature representing the loaded geometry information, with the given properties
        """
        geojson_geometry_str = to_geojson(self.shape)
        geojson_geometry = json.loads(geojson_geometry_str)
        geojson_feature = {
            "type": "Feature",
            "geometry": geojson_geometry,
            "properties": properties
        }

        return geojson_feature

    @staticmethod
    def transform_crs(shp, from_crs):
        """
        Function that transforms the CRS (Coordinate Reference System) of the Shapely shape to WGS84 (EPSG:4326)

        :shp: input Shapely shape in any CRS that is supported by pyproj
        :from_crs: the CRS used for the Shapely shape
        :return: Shapely shape transformed to WGS84
        """
        if from_crs == "WGS84" or from_crs == "" or from_crs is None:
            from_crs = "EPSG:4326"

        if from_crs != "EPSG:4326":
            wgs84 = pyproj.CRS("EPSG:4326")
            original_crs = pyproj.CRS(from_crs)

            project = pyproj.Transformer.from_crs(original_crs, wgs84, always_xy=True).transform
            return transform(project, shp)
        else:
            return shp


class ShapePoint(Shape):
    """
    Represents a point. Provides functionality to convert to and from esdl.Point, Shapely Point, WKT/WKB,
    Leaflet (used in the ESDL MapEditor), GeoJSON. Uses a Shapely Point internally and provides different methods
    to convert to and from different other formats.
    """

    def __init__(self, shape_input):
        """
        Constructor of the ShapePoint class. Can be called with an esdl.Point, a dictionary with "lat" and "lng" keys,
        or a Shapely Point instance

        :param shape_input: the input shape, can be an esdl.Point, a dictionary with "lat" and "lng" keys,
                            or a Shapely Point instance
        """
        if isinstance(shape_input, esdl.Point):
            self.shape = self.parse_esdl(shape_input)
        elif isinstance(shape_input, dict) and "lat" in shape_input:
            self.shape = self.parse_leaflet(shape_input)
        elif isinstance(shape_input, Point):
            self.shape = shape_input
        else:
            raise Exception("ShapePoint constructor called with unsupported type")

    @staticmethod
    def parse_esdl(esdl_geometry):
        """
        Function that uses an esdl.Point instance as an input to generate a Shapely Point (using WGS84 as CRS)

        :param esdl_geometry: an esdl.Point instance
        :return: a Shapely shape (using WGS84 as CRS)
        """
        if isinstance(esdl_geometry, esdl.Point):
            return Shape.transform_crs(Point(esdl_geometry.lon, esdl_geometry.lat), esdl_geometry.CRS)
        else:
            raise Exception("Cannot instantiate a Shapely Point with an ESDL geometry other than esdl.Point")

    @staticmethod
    def parse_leaflet(leaflet_coords):
        """
        Function that uses a dictionary with "lat" and "lng" keys as an input to generate a Shapely Point (using
        WGS84 as CRS)

        :param esdl_geometry: a dictionary with "lat" and "lng" keys
        :return: a Shapely Point (using WGS84 as CRS)
        """
        if isinstance(leaflet_coords, dict) and "lat" in leaflet_coords:
            return Point(leaflet_coords["lng"], leaflet_coords["lat"])
        else:
            raise Exception("Incorrect instantiation of a Shapely Point with leaflet coordinates")

    def get_esdl(self):
        """
        Function that generates an esdl.Point instance based on the loaded shape.

        :return: an esdl.Point instance based on the loaded shape
        """
        return esdl.Point(lon=self.shape.coords[0][0], lat=self.shape.coords[0][1])


class ShapeLine(Shape):
    """
    Represents a line. Provides functionality to convert to and from esdl.Line, Shapely LineString, WKT/WKB,
    Leaflet (used in the ESDL MapEditor), GeoJSON. Uses a Shapely LineString internally and provides different methods
    to convert to and from different other formats.
    """

    def __init__(self, shape_input):
        """
        Constructor of the ShapeLine class. Can be called with an esdl.Line, a list of dictionaries with "lat" and
        "lng" keys, or a Shapely LineString instance

        :param shape_input: the input shape, can be an esdl.Line, a list of dictionaries with "lat" and
                            "lng" keys, or a Shapely LineString instance
        """
        if isinstance(shape_input, esdl.Line):
            self.shape = self.parse_esdl(shape_input)
        elif isinstance(shape_input, list) and all(isinstance(elem, dict) for elem in shape_input):
            self.shape = self.parse_leaflet(shape_input)
        elif isinstance(shape_input, LineString):
            self.shape = shape_input
        else:
            raise Exception("ShapeLine constructor called with unsupported type")

    @staticmethod
    def parse_esdl(esdl_geometry):
        """
        Function that uses an esdl.Line instance as an input to generate a Shapely LineString (using WGS84 as CRS)

        :param esdl_geometry: an esdl.Line instance
        :return: a Shapely LineString (using WGS84 as CRS)
        """
        if isinstance(esdl_geometry, esdl.Line):
            linestring = list()
            for p in esdl_geometry.point:
                linestring.append((p.lon, p.lat))
            return Shape.transform_crs(LineString(linestring), esdl_geometry.CRS)
        else:
            raise Exception("Cannot instantiate a Shapely LineString with an ESDL geometry other than esdl.Line")

    @staticmethod
    def parse_leaflet(leaflet_coords):
        """
        Function that uses a list of dictionaries with "lat" and "lng" keys as an input to generate a Shapely
        LineString (using WGS84 as CRS)

        :param esdl_geometry: a list of dictionaries with "lat" and "lng" keys
        :return: a Shapely LineString (using WGS84 as CRS)
        """
        if isinstance(leaflet_coords, list) and all(isinstance(elem, dict) for elem in leaflet_coords):
            linestring = list()
            for elem in leaflet_coords:
                linestring.append((elem["lng"], elem["lat"]))
            return LineString(linestring)
        else:
            raise Exception("Incorrect instantiation of a Shapely LineString with leaflet coordinates")

    def get_esdl(self):
        """
        Function that generates an esdl.Line instance based on the loaded shape.

        :return: an esdl.Line instance based on the loaded shape
        """
        line = esdl.Line()
        for p in self.shape.coords:
            point = esdl.Point(lon=p[0], lat=p[1])
            line.point.append(point)
        return line


class ShapePolygon(Shape):
    """
    Represents a polygon. Provides functionality to convert to and from esdl.Polygon, Shapely Polygon, WKT/WKB,
    Leaflet (used in the ESDL MapEditor), GeoJSON. Uses a Shapely Polygon internally and provides different methods
    to convert to and from different other formats.
    """

    def __init__(self, shape_input):
        """
        Constructor of the ShapePolygon class. Can be called with an esdl.Polygon, a list of lists of dictionaries
        with "lat" and "lng" keys, or a Shapely Polygon instance

        :param shape_input: the input shape, can be an esdl.Polygon, a list of lists of dictionaries
                            with "lat" and "lng" keys, or a Shapely Polygon instance
        """
        if isinstance(shape_input, esdl.Polygon):
            self.shape = self.parse_esdl(shape_input)
        elif isinstance(shape_input, list) and all(isinstance(elem, list) for elem in shape_input):
            self.shape = self.parse_leaflet(shape_input)
        elif isinstance(shape_input, Polygon):
            self.shape = shape_input
        else:
            raise Exception("ShapePolygon constructor called with unsupported type")

    @staticmethod
    def parse_esdl(esdl_geometry):
        """
        Function that uses an esdl.Polygon instance as an input to generate a Shapely Polygon (using WGS84 as CRS)

        :param esdl_geometry: an esdl.Polygon instance
        :return: a Shapely Polygon (using WGS84 as CRS)
        """
        if isinstance(esdl_geometry, esdl.Polygon):
            exterior = list()
            interiors = list()
            for p in esdl_geometry.exterior.point:
                exterior.append([p.lon, p.lat])
            for pol in esdl_geometry.interior:
                interior = list()
                for p in pol.point:
                    interior.append([p.lon, p.lat])
                interiors.append(interior)
            return Shape.transform_crs(Polygon(exterior, interiors), esdl_geometry.CRS)
        else:
            raise Exception("Cannot instantiate a Shapely Polygon with an ESDL geometry other than esdl.Polygon")

    @staticmethod
    def parse_leaflet(leaflet_coords):
        """
        Function that uses a list of lists of dictionaries with "lat" and "lng" keys as an input to generate a Shapely
        Polygon (using WGS84 as CRS). The first element in the list represents the exterior of the polygon, the other
        elements represent zero, one or more interiors (holes in the polygon)

        :param esdl_geometry: a list of lists of dictionaries with "lat" and "lng" keys
        :return: a Shapely Polygon (using WGS84 as CRS)
        """
        if isinstance(leaflet_coords, list) and all(isinstance(elem, list) for elem in leaflet_coords):
            exterior = list()
            interiors = list()
            lc_exterior = leaflet_coords.pop(0)
            for p in lc_exterior:
                exterior.append([p["lng"], p["lat"]])
            for pol in leaflet_coords:
                interior = list()
                for p in pol:
                    interior.append([p["lng"], p["lat"]])
                interiors.append(interior)
            return Polygon(exterior, interiors)
        else:
            raise Exception("Incorrect instantiation of a Shapely Polygon with leaflet coordinates")

    def get_esdl(self):
        """
        Function that generates an esdl.Polygon instance based on the loaded shape.

        :return: an esdl.Polygon instance based on the loaded shape
        """
        pol = esdl.Polygon()
        exterior = esdl.SubPolygon()
        for p in self.shape.exterior.coords:
            exterior.point.append(esdl.Point(lon=p[0], lat=p[1]))
        pol.exterior = exterior
        for ip in self.shape.interiors:
            interior = esdl.SubPolygon()
            for p in ip.coords:
                interior.point.append(esdl.Point(lon=p[0], lat=p[1]))
            pol.interior.append(interior)
        return pol


class ShapeMultiPolygon(Shape):
    """
    Represents a multipolygon. Provides functionality to convert to and from esdl.MultiPolygon, Shapely MultiPolygon,
    WKT/WKB, Leaflet (used in the ESDL MapEditor), GeoJSON. Uses a Shapely MultiPolygon internally and provides
    different methods to convert to and from different other formats.
    """

    def __init__(self, shape_input):
        """
        Constructor of the ShapeMultiPolygon class. Can be called with an esdl.MultiPolygon, a list of lists of lists
        of dictionaries with "lat" and "lng" key, or a Shapely Polygon instance

        :param shape_input: the input shape, can be an esdl.Polygon, a list of lists of dictionaries
                            with "lat" and "lng" key, or a Shapely Polygon instance
        """
        if isinstance(shape_input, esdl.MultiPolygon):
            self.shape = self.parse_esdl(shape_input)
        elif isinstance(shape_input, list) and all(isinstance(elem, list) for elem in shape_input):
            self.shape = self.parse_leaflet(shape_input)
        elif isinstance(shape_input, MultiPolygon):
            self.shape = shape_input
        else:
            raise Exception("ShapeMultiPolygon constructor called with unsupported type")

    @staticmethod
    def parse_esdl(esdl_geometry):
        """
        Function that uses an esdl.MultiPolygon instance as an input to generate a Shapely MultiPolygon (using WGS84
        as CRS)

        :param esdl_geometry: an esdl.MultiPolygon instance
        :return: a Shapely MultiPolygon (using WGS84 as CRS)
        """
        if isinstance(esdl_geometry, esdl.MultiPolygon):
            plist = list()
            for p in esdl_geometry.polygon:
                plist.append(ShapePolygon.parse_esdl(p))
            return Shape.transform_crs(MultiPolygon(plist), esdl_geometry.CRS)
        else:
            raise Exception(
                "Cannot instantiate a Shapely MultiPolygon with an ESDL geometry other than esdl.MultiPolygon")

    @staticmethod
    def parse_leaflet(leaflet_coords):
        """
        Function that uses a list of lists of lists of dictionaries with "lat" and "lng" keys as an input to generate
        a Shapely MultiPolygon (using WGS84 as CRS). Iterates over the outer list and calls ShapePolygon.parse_leaflet
        for each element

        :param esdl_geometry: a list of lists of lists of dictionaries with "lat" and "lng" keys
        :return: a Shapely MultiPolygon (using WGS84 as CRS)
        """
        if isinstance(leaflet_coords, list) and all(isinstance(elem, list) for elem in leaflet_coords):
            plist = list()
            for p in leaflet_coords:
                plist.append(ShapePolygon.parse_leaflet(p))
            return MultiPolygon(plist)
        else:
            raise Exception(
                "Incorrect instantiation of a Shapely MultiPolygon with leaflet coordinates")

    def get_esdl(self):
        """
        This get_esdl function is not yet implemented for the ShapeMultiPolygon class, as MultiPolygon
        is not a frequent ESDL geometry
        """
        raise Exception("Not implemented yet, MultiPolygon is not a frequent ESDL geometry")

    def get_polygon_list_esdl(self):
        """
        Function that generates a list of esdl.Polygons, based on the loaded MultiPolygon shape.

        :return: list of esdl.Polygons
        """
        pol_list = list()

        for mp_element in self.shape.geoms:
            polygon = ShapePolygon(mp_element)
            pol_list.append(polygon.get_esdl())

        return pol_list


class ShapeGeometryCollection(Shape):
    """
    Represents a geometry collection. Provides functionality to convert to and from Shapely GeometryCollection,
    WKT/WKB and GeoJSON. Uses a Shapely GeometryCollection internally and provides different methods
    to convert to and from different other formats.
    """

    def __init__(self, shape_input):
        """
        Constructor of the ShapeGeometryCollection class. Can be called with a Shapely GeometryCollection instance

        :param shape_input: the input shape, can be a Shapely GeometryCollection instance
        """
        if isinstance(shape_input, GeometryCollection):
            self.shape = shape_input
        else:
            raise Exception("ShapeMultiPolygon constructor called with unsupported type")

    def get_esdl(self):
        """
        This get_esdl function is not yet implemented for the ShapeGeometryCollection class, as GeometryCollection
        is not a frequent ESDL geometry
        """
        raise Exception("Not implemented yet, GeometryCollection is not a frequent ESDL geometry")



"""
Temporary replace xmi.py from pyEcore with our own as multi-valued attributed values are not correctly read from the
xmi file. In pyEcore 0.13 + 0.14 they used () tuple notation instead of [] for multi-valued attribute values causing it to return a
generator for those values instead of a list of values.

The xmi module introduces XMI resource and XMI parsing.
"""
from enum import unique, Enum
from functools import lru_cache
from lxml.etree import parse, QName, Element, SubElement, ElementTree
from pyecore.resources.resource import Resource
from pyecore.ecore import EClass, EStringToStringMapEntry, EAnnotation, EProxy, \
                    EDataType

XSI = 'xsi'
XSI_URL = 'http://www.w3.org/2001/XMLSchema-instance'
XMI = 'xmi'
XMI_URL = 'http://www.omg.org/XMI'


@unique
class XMIOptions(Enum):
    OPTION_USE_XMI_TYPE = 0
    SERIALIZE_DEFAULT_VALUES = 1


class XMIResource(Resource):
    def __init__(self, uri=None, use_uuid=False):
        super().__init__(uri, use_uuid)
        self._later = []
        self.prefixes = {}
        self.reverse_nsmap = {}

    def load(self, options=None):
        self.options = options or {}
        self.cache_enabled = True
        tree = parse(self.uri.create_instream())
        xmlroot = tree.getroot()
        self.prefixes.update(xmlroot.nsmap)
        self.reverse_nsmap = {v: k for k, v in self.prefixes.items()}

        self.xsitype = f'{{{self.prefixes.get(XSI)}}}type'
        self.xmiid = f'{{{self.prefixes.get(XMI)}}}id'
        self.schema_tag = f'{{{self.prefixes.get(XSI)}}}schemaLocation'

        # Decode the XMI
        if f'{{{self.prefixes.get(XMI)}}}XMI' == xmlroot.tag:
            real_roots = xmlroot
        else:
            real_roots = [xmlroot]

        def grouper(iterable):
            args = [iter(iterable)] * 2
            return zip(*args)

        self.schema_locations = {}
        schema_tag_list = xmlroot.attrib.get(self.schema_tag, '')
        for prefix, path in grouper(schema_tag_list.split()):
            if '#' not in path:
                path = path + '#'
            self.schema_locations[prefix] = EProxy(path, self)

        for root in real_roots:
            modelroot = self._init_modelroot(root)
            for child in root:
                self._decode_eobject(child, modelroot)

        if self.contents:
            self._decode_ereferences()

        self._clean_registers()
        self.uri.close_stream()

    def xsi_type_url(self):
        if self.options.get(XMIOptions.OPTION_USE_XMI_TYPE, False):
            return XMI_URL
        return XSI_URL

    @staticmethod
    def extract_namespace(tag):
        qname = QName(tag)
        return qname.namespace, qname.localname

    def _type_attribute(self, node):
        type_ = node.get(self.xsitype)
        if type_ is None:
            xmi_type_url = f'{{{self.prefixes.get(XMI)}}}type'
            type_ = node.get(xmi_type_url)
        return type_

    def _get_metaclass(self, nsURI, eclass_name):
        try:
            return self.get_metamodel(nsURI).getEClassifier(eclass_name)
        except Exception as e:
            proxy = self.schema_locations[nsURI]
            try:
                return proxy.getEClassifier(eclass_name)
            except Exception:
                raise e

    def _init_modelroot(self, xmlroot):
        nsURI, eclass_name = self.extract_namespace(xmlroot.tag)
        eclass = self._get_metaclass(nsURI, eclass_name)
        if not eclass:
            raise TypeError(f'"{eclass_name}" EClass does not exists '
                            f'(line {xmlroot.tag})')
        modelroot = eclass()
        modelroot._eresource = self
        self.use_uuid = xmlroot.get(self.xmiid) is not None
        self.contents.append(modelroot)
        erefs = []
        for key, value in xmlroot.attrib.items():
            namespace, _ = self.extract_namespace(key)
            if key == self.xmiid:
                modelroot._internal_id = value
                self.uuid_dict[value] = modelroot
            # Do stuff with this
            # elif namespace:
            #     try:
            #
            #         # metaclass = self.get_metamodel(namespace)
            #         pass
            #     except KeyError:
            #         pass
            elif not namespace:
                feature = self._find_feature(modelroot.eClass, key)
                if not feature:
                    continue
                if feature.is_attribute:
                    self._decode_eattribute_value(modelroot, feature, value)
                    if feature.iD:
                        self.uuid_dict[value] = modelroot
                else:
                    erefs.append((feature, value))
        if erefs:
            self._later.append((modelroot, erefs))
        return modelroot

    @staticmethod
    def _decode_eattribute_value(eobject, eattribute, value, from_tag=False):
        is_many = eattribute.many
        if is_many and not from_tag:
            values = value.split()
            from_string = eattribute._eType.from_string
            results = [from_string(x) for x in values]
            eobject.__getattribute__(eattribute.name).extend(results)
        elif is_many:
            value = eattribute._eType.from_string(value)
            eobject.__getattribute__(eattribute._name).append(value)
        else:
            val = eattribute._eType.from_string(value)
            eobject.__setattr__(eattribute._name, val)

    def _decode_eobject(self, current_node, parent_eobj):
        eobject_info = self._decode_node(parent_eobj, current_node)
        feat_container, eobject, eatts, erefs, from_tag = eobject_info

        # deal with eattributes and ereferences
        for eattribute, value in eatts:
            self._decode_eattribute_value(eobject, eattribute, value, from_tag)

        if erefs:
            self._later.append((eobject, erefs))

        if not feat_container:
            return

        # attach the new eobject to the parent one
        if feat_container.many:
            parent_eobj.__getattribute__(feat_container._name).append(eobject)
        else:
            parent_eobj.__setattr__(feat_container._name, eobject)

        # iterate on children
        for child in current_node:
            self._decode_eobject(child, eobject)

    def _is_none_node(self, node):
        return f'{{{XSI_URL}}}nil' in node.attrib

    def _decode_node(self, parent_eobj, node):
        _, node_tag = self.extract_namespace(node.tag)
        feature_container = self._find_feature(parent_eobj.eClass, node_tag)
        if not feature_container:
            raise ValueError(f'Feature "{node_tag}" is unknown '
                             f'for {parent_eobj.eClass.name}, '
                             f'line {node.sourceline}')
        if self._is_none_node(node):
            if feature_container.many:
                parent_eobj.__getattribute__(feature_container._name) \
                           .append(None)
            else:
                parent_eobj.__setattr__(feature_container._name, None)

            return (None, None, [], [], False)
        if node.get('href'):
            ref = node.get('href')
            proxy = EProxy(path=ref, resource=self)
            return (feature_container, proxy, [], [], False)
        if self._type_attribute(node):
            prefix, _type = self._type_attribute(node).split(':')
            if not prefix:
                raise ValueError(f'Prefix {prefix} is not registered, '
                                 f'{node.tag} line {node.sourceline}')
            epackage = self.prefix2epackage(prefix)
            etype = epackage.getEClassifier(_type)
            if not etype:
                raise ValueError(f'Type {_type} is unknown in {epackage}, '
                                 f'{node.tag} line {node.sourceline}')
        else:
            etype = feature_container._eType
            if isinstance(etype, EProxy):
                etype.force_resolve()

        # we create the instance
        if etype is EClass or etype is EClass.eClass:
            name = node.get('name')
            eobject = etype(name)
        elif (etype is EStringToStringMapEntry
              or etype is EStringToStringMapEntry.eClass) \
                and feature_container is EAnnotation.details:
            annotation_key = node.get('key')
            annotation_value = node.get('value')
            parent_eobj.details[annotation_key] = annotation_value
            if annotation_key == 'documentation':
                container = parent_eobj.eContainer()
                if hasattr(container, 'python_class'):
                    container = container.python_class
                container.__doc__ = annotation_value
            return (None, None, (), (), False)
        elif isinstance(etype, EDataType):
            value = node.text if node.text else ''
            return (None, parent_eobj, ((feature_container, value),),
                    (), True)
        else:
            # idref = node.get(f'{{{XMI_URL}}}idref')
            # if idref:
            #     return (None, parent_eobj, [],
            #             [(feature_container, idref)], True)
            eobject = etype()

        # we sort the node feature (no containments)
        eatts = []
        erefs = []
        for key, value in node.attrib.items():
            feature = self._decode_attribute(eobject, key, value, node)
            if not feature:
                continue  # we skip the unknown features
            if etype is EClass and feature._name == 'name':
                continue  # we skip the name for metamodel import
            if feature.is_attribute:
                eatts.append((feature, value))
                if feature.iD:
                    self.uuid_dict[value] = eobject
            else:
                erefs.append((feature, value))
        return (feature_container, eobject, eatts, erefs, False)

    def _decode_attribute(self, owner, key, value, node):
        namespace, att_name = self.extract_namespace(key)
        prefix = self.reverse_nsmap[namespace] if namespace else None
        # This is a special case, we are working with uuids
        if key == self.xmiid:
            owner._internal_id = value
            self.uuid_dict[value] = owner
        elif prefix in ('xsi', 'xmi') and att_name == 'type':
            # type has already been handled
            pass
        # elif namespace:
        #     pass
        elif not namespace:
            if att_name == 'href':
                return
            feature = self._find_feature(owner.eClass, att_name)
            if not feature:
                raise ValueError(f'Feature {att_name} does not exists for '
                                 f'type {owner.eClass.name} '
                                 f'({node.tag} line {node.sourceline})')
            return feature

    def _decode_ereferences(self):
        opposite = []
        for eobject, erefs in self._later:
            for ref, value in erefs:
                name = ref._name
                if name == 'eOpposite':
                    opposite.append((eobject, ref, value))
                    continue
                if ref.many:
                    values = [self.normalize(x) for x in value.split()]
                else:
                    values = [value]
                for value in values:
                    if not value:  # BA: Skip empty references
                        continue
                    resolved_value = self._resolve_nonhref(value)
                    if not resolved_value:
                        raise ValueError(f'EObject for {value} is unknown')
                    if not hasattr(resolved_value, '_inverse_rels'):
                        resolved_value = resolved_value.eClass
                    if ref.many:
                        eobject.__getattribute__(name).append(resolved_value)
                    else:
                        eobject.__setattr__(name, resolved_value)

        for eobject, ref, value in opposite:
            resolved_value = self._resolve_nonhref(value)
            if not resolved_value:
                raise ValueError(f'EObject for {value} is unknown')
            eobject.__setattr__(ref._name, resolved_value)

    @lru_cache()
    def _resolve_nonhref(self, path):
        uri, fragment = self._is_external(path)
        if uri:
            cleaned_uri = uri + '#' + fragment
            if cleaned_uri in self._resolve_mem:
                return self._resolve_mem[cleaned_uri]
            proxy = EProxy(path=cleaned_uri, resource=self)
            self._resolve_mem[cleaned_uri] = proxy
            return proxy
        if fragment in self._resolve_mem:
            return self._resolve_mem[fragment]
        return self.resolve(fragment)

    def _clean_registers(self):
        self._later.clear()
        self._find_feature.cache_clear()
        self._resolve_nonhref.cache_clear()
        self._resolve_mem.clear()
        self.cache_enabled = False

    def register_nsmap(self, prefix, uri):
        if uri in self.reverse_nsmap:
            return
        if prefix not in self.prefixes:
            self.prefixes[prefix] = uri
            self.reverse_nsmap[uri] = prefix
            return
        same_prefix = [x for x in self.prefixes.keys() if x.startswith(prefix)]
        prefix = f'{prefix}_{len(same_prefix)}'
        self.prefixes[prefix] = uri
        self.reverse_nsmap[uri] = prefix

    def register_eobject_epackage(self, eobj):
        epackage = eobj.eClass.ePackage
        prefix = epackage.nsPrefix
        nsURI = epackage.nsURI
        self.register_nsmap(prefix, nsURI)

    def save(self, output=None, options=None):
        self.options = options or {}
        output = self.open_out_stream(output)
        self.prefixes.clear()
        self.reverse_nsmap.clear()

        serialize_default = \
            self.options.get(XMIOptions.SERIALIZE_DEFAULT_VALUES,
                             False)
        nsmap = {XMI: XMI_URL}

        if len(self.contents) == 1:
            root = self.contents[0]
            self.register_eobject_epackage(root)
            tmp_xmi_root = self._go_across(root, serialize_default)
        else:
            tag = QName(XMI_URL, 'XMI')
            tmp_xmi_root = Element(tag)
            for root in self.contents:
                root_node = self._go_across(root, serialize_default)
                tmp_xmi_root.append(root_node)

        # update nsmap with prefixes register during the nodes creation
        nsmap.update(self.prefixes)
        xmi_root = Element(tmp_xmi_root.tag, nsmap=nsmap)
        xmi_root[:] = tmp_xmi_root[:]
        xmi_root.attrib.update(tmp_xmi_root.attrib)
        xmi_version = QName(XMI_URL, 'version')
        xmi_root.attrib[xmi_version] = '2.0'
        tree = ElementTree(xmi_root)
        tree.write(output,
                   pretty_print=True,
                   xml_declaration=True,
                   encoding=tree.docinfo.encoding)
        output.flush()
        self.uri.close_stream()

    def _add_explicit_type(self, node, obj):
        self.prefixes[XSI] = XSI_URL
        xsi_type = QName(self.xsi_type_url(), 'type')
        uri = obj.eClass.ePackage.nsURI
        if uri not in self.reverse_nsmap:
            # epackage = self.get_metamodel(uri)
            epackage = obj.eClass.ePackage
            self.register_nsmap(epackage.nsPrefix, uri)
        prefix = self.reverse_nsmap[uri]
        node.attrib[xsi_type] = f'{prefix}:{obj.eClass.name}'

    def _build_none_node(self, feature_name):
        sub = Element(feature_name)
        xsi_null = QName(self.xsi_type_url(), 'nil')
        sub.attrib[xsi_null] = 'true'
        return sub

    def _go_across(self, obj, serialize_default=False):
        eclass = obj.eClass
        if obj.eContainmentFeature():
            node = Element(obj.eContainmentFeature()._name)
            if obj.eContainmentFeature()._eType != eclass:
                self._add_explicit_type(node, obj)
        else:
            # obj is the root
            epackage = eclass.ePackage
            nsURI = epackage.nsURI
            tag = QName(nsURI, eclass.name) if nsURI else eclass.name
            node = Element(tag)

        if self.use_uuid:
            self._assign_uuid(obj)
            xmi_id = f'{{{XMI_URL}}}id'
            node.attrib[xmi_id] = obj._internal_id

        for feat in obj._isset:
            if feat.derived or feat.transient:
                continue
            feat_name = feat._name
            value = obj.__getattribute__(feat_name)
            if value is None:
                if serialize_default:
                    node.append(self._build_none_node(feat_name))
                continue
            if hasattr(feat._eType, 'eType') and feat._eType.eType is dict:
                for key, val in value.items():
                    entry = Element(feat_name)
                    entry.attrib['key'] = key
                    entry.attrib['value'] = val
                    node.append(entry)
            elif feat.is_attribute:
                etype = feat._eType
                if feat.many and value:
                    to_str = etype.to_string
                    has_special_char = False
                    result_list = []
                    for v in value:
                        string = None if v is None else to_str(v)
                        if not string or any(x.isspace() for x in string):
                            has_special_char = True
                        result_list.append(string)
                    if has_special_char:
                        for v in result_list:
                            if v is None:
                                node.append(self._build_none_node(feat_name))
                            else:
                                sub = SubElement(node, feat_name)
                                sub.text = v
                    else:
                        node.attrib[feat_name] = ' '.join(result_list)
                    continue
                default_value = feat.get_default_value()
                if value != default_value or serialize_default:
                    node.attrib[feat_name] = etype.to_string(value)
                continue

            elif feat.eOpposite and feat.eOpposite.containment:
                continue
            elif not feat.containment:
                if feat.many:
                    results = (self._build_path_from(x) for x in value)
                    embedded = []
                    crossref = []
                    for i, result in enumerate(results):
                        frag, cref = result
                        if cref:
                            crossref.append((i, frag))
                        else:
                            embedded.append(frag)
                    if embedded:
                        result = ' '.join(embedded)
                        node.attrib[feat_name] = result
                    for i, ref in crossref:
                        sub = SubElement(node, feat_name)
                        sub.attrib['href'] = ref
                        self._add_explicit_type(sub, value[i])
                else:
                    frag, is_crossref = self._build_path_from(value)
                    if is_crossref:
                        sub = SubElement(node, feat_name)
                        sub.attrib['href'] = frag
                        self._add_explicit_type(sub, value)
                    else:
                        node.attrib[feat_name] = frag

            else:
                children = value if feat.many else [value]
                for child in children:
                    node.append(self._go_across(child, serialize_default))
        return node


#  This work is based on original code developed and copyrighted by TNO 2023.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO
import copy
import uuid
from typing import Union

from esdl import esdl, PhysicalQuantityEnum
from esdl.units.conversion import equals

unitdict = {
    'NONE': '-',
    'AMPERE': 'A',
    'JOULE': 'J',
    'WATTHOUR': 'Wh',
    'WATT': 'W',
    'VOLT': 'V',
    'BAR': 'bar',
    'PSI': 'psi',
    'DEGREES_CELSIUS': '\u2103',  # Sign for degrees Celcius
    'KELVIN': 'K',
    'GRAM': 'g',
    'TONNE': 't',
    'EURO': 'EUR',
    'DOLLAR': 'USD',
    'METRE': 'm',
    'SQUARE_METRE': 'm2',
    'CUBIC_METRE': 'm3',
    'LITRE': 'l',
    'WATTSECOND': 'Ws',
    'ARE': 'a',
    'HECTARE': 'ha',
    'PERCENT': '%',
    'VOLT_AMPERE': 'VA',
    'VOLT_AMPERE_REACTIVE': 'VAR',
    'PASCAL': 'Pa',
    'NEWTON': 'N',
    'DEGREES': '\u00b0',  # Sign for degrees
    'HOUR': 'h'
}


timeunitdict = {
    'SECOND': 'sec',
    'MINUTE': 'min',
    'QUARTER': '15mins',
    'HOUR': 'hr',
    'DAY': 'day',
    'WEEK': 'wk',
    'MONTH': 'mon',
    'YEAR': 'yr'
}


multiplierdict = {
    'ATTO': 'a',
    'FEMTO': 'f',
    'PICO': 'p',
    'NANO': 'n',
    'MICRO': 'u',
    'MILLI': 'm',
    'KILO': 'k',
    'MEGA': 'M',
    'GIGA': 'G',
    'TERA': 'T',
    'TERRA': 'T',       # due to spelling mistake in ESDL
    'PETA': 'P',
    'EXA': 'E'
}


def qau_to_string(qau):
    """
    Converts a QuantityAndUnit instance to a string, for example "POWER in MW".

    :param qau: an esdl.QuantityAndUnit instance
    :result: string representation of the QuanityAndUnit instance
    """
    s = qau.physicalQuantity.name
    str_unit = unit_to_string(qau)
    if str_unit != '':
        s += ' in ' + str_unit

    return s


def unit_to_string(qau):
    """
    Converts the unit of a QuantityAndUnit instance to a string, for example "MW".

    :param qau: an esdl.QuantityAndUnit instance
    :result: string representation of the unit only of the QuanityAndUnit instance
    """
    mult = qau.multiplier.name
    unit = qau.unit.name
    pmult = qau.perMultiplier.name
    punit = qau.perUnit.name
    ptunit = qau.perTimeUnit.name

    s = ''

    if unit != 'NONE' and unit != 'UNDEFINED':
        if mult != 'NONE' and mult != 'UNDEFINED':
            s += multiplierdict[mult]
        try:
            s += unitdict[unit]
        except KeyError:
            s += unit
    if punit != 'NONE' and punit != 'UNDEFINED':
        s += '/'
        if pmult != 'NONE' and pmult != 'UNDEFINED':
            s += multiplierdict[pmult]
        try:
            s += unitdict[punit]
        except KeyError:  # SECOND etc is not in the dict
            s += punit
    if ptunit != 'NONE' and ptunit != 'UNDEFINED':
        s += '/' + timeunitdict[ptunit]

    return s


def build_qau_from_unit_string(unit_string: str, physical_quantity: esdl.PhysicalQuantityEnum = None):
    """
    Build an esdl.QuantityAndUnit instance from a string representing only the unit (and not the physical quantity),
    for example from "kWh/yr".

    :param unit_string: string representation of the QuanityAndUnit unit (without the physical quantity)
    :param physical_quantity: Optional sets the associated Physical quantity, e.g. esdl.PhysicalQuantityEnum.POWER
           or esdl.PhysicalQuantityEnum.ENERGY, esdl.PhysicalQuantityEnum.COST, also string version is supported, e.g.
           "Power", "Energy", "Cost" (case does not matter), but Python 3.10 does not like the
           esdl.PhysicalQuantityEnum | str definition in the method definition.
    :result: an esdl.QuantityAndUnit instance
    """

    qau = esdl.QuantityAndUnitType(id=str(uuid.uuid4()))

    if physical_quantity:
        if isinstance(physical_quantity, esdl.PhysicalQuantityEnum):
            qau.physicalQuantity = physical_quantity
        elif isinstance(physical_quantity, str):
            pq = esdl.PhysicalQuantityEnum.from_string(physical_quantity.upper())
            if pq:
                qau.physicalQuantity = pq

    unit_parts = unit_string.split('/')
    if unit_parts:
        # Parse the unit
        for u in unitdict:
            if unitdict[u] == unit_parts[0]:
                qau.unit = esdl.UnitEnum.from_string(u)
                break

        # if the first try failed, try to see if there is a multiplier in front of the unit
        if qau.unit == esdl.UnitEnum.NONE:
            unit = unit_parts[0][1:]
            for u in unitdict:
                if unitdict[u] == unit:
                    for m in multiplierdict:
                        if multiplierdict[m] == unit_parts[0][0]:
                            qau.unit = esdl.UnitEnum.from_string(u)
                            qau.multiplier = esdl.MultiplierEnum.from_string(m)
                            break
                    break

        # Zero, one or two 'perUnits' are possible
        if len(unit_parts) > 1:
            for up in range(1, len(unit_parts)):
                # Parse the perUnit
                for u in unitdict:
                    if unitdict[u] == unit_parts[up]:
                        qau.perUnit = esdl.UnitEnum.from_string(u)
                        break

                # if the first try failed, try to see if there is a multiplier in front of the perUnit
                if qau.perUnit == esdl.UnitEnum.NONE:
                    unit = unit_parts[up][1:]
                    for u in unitdict:
                        if unitdict[u] == unit:
                            for m in multiplierdict:
                                if multiplierdict[m] == unit_parts[up][0]:
                                    qau.perUnit = esdl.UnitEnum.from_string(u)
                                    qau.perMultiplier = esdl.MultiplierEnum.from_string(m)
                                    break
                            break

                # Parse the perTimeUnit
                for tu in timeunitdict:
                    if timeunitdict[tu] == unit_parts[up]:
                        qau.perTimeUnit = esdl.TimeUnitEnum.from_string(tu)
                        break

    return qau


def is_valid_uuid(uuid_to_test, version=4):
    """
    Check if uuid_to_test is a valid UUID.

    Parameters
    ----------
    uuid_to_test : str
    version : {1, 2, 3, 4}

    Returns
    -------
    `True` if uuid_to_test is a valid UUID, otherwise `False`.

    Examples
    --------
    >>> is_valid_uuid('c9bf9e57-1685-4c89-bafb-ff5af830be8a')
    True
    >>> is_valid_uuid('c9bf9e58')
    False
    """

    try:
        uuid_obj = uuid.UUID(uuid_to_test, version=version)
    except ValueError:
        return False
    return str(uuid_obj) == uuid_to_test


def instantiate_qau(orig_qau: esdl.QuantityAndUnitType):
    """
    Creates a new instance of a QuantityAndUnit instance by copying the contents and creating a new ID.

    :param orig_qau: an esdl.QuantityAndUnit instance used as template
    """
    qau = orig_qau.deepcopy()
    qau.id = str(uuid.uuid4())
    return qau


def get_or_create_global_qau_reference(es: esdl.EnergySystem, orig_qau: esdl.QuantityAndUnitType):
    """
    Finds a QuantityAndUnit instance in the global list of QuantityAndUnit instances in the EnergySystemInformation.
    It the QuantityAndUnit exists, a QuantityAndUnitReference is returned. If the QuantityAndUnit does not exist,
    it will be created and a QuantityAndUnitReference is returned. Must be used with the default QuantityAndUnits, like
    defined in conversion.py, like ENERGY_IN_PJ, COST_IN_MEur

    :param es: the EnergySystem instance
    :param orig_qau: a QuantityAndUnit instance from the list of QuantityAndUnit templates
    """
    esi = es.energySystemInformation
    if not esi:
        esi = es.energySystemInformation = esdl.EnergySystemInformation()

    qaus = esi.quantityAndUnits
    if not qaus:
        qaus = esi.quantityAndUnits = esdl.QuantityAndUnits(id=str(uuid.uuid4()), name="Global quantities and units")

    qau_ref = None
    for qau in qaus.quantityAndUnit:
        if equals(qau, orig_qau):
            qau_ref = esdl.QuantityAndUnitReference(reference=qau)

    if not qau_ref:
        new_qau = orig_qau.deepcopy()
        if is_valid_uuid(new_qau.id, 4):
            new_qau.id = str(uuid.uuid4())     # If the orig_qau had an UUID as ID, change it

        qaus.quantityAndUnit.append(new_qau)
        qau_ref = esdl.QuantityAndUnitReference(reference=new_qau)

    return qau_ref



#  This work is based on original code developed and copyrighted by TNO 2023.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO

import codecs
import csv
import locale
from datetime import datetime, timedelta
from enum import Enum
from uuid import uuid4

import pytz
from pyecore.ecore import EDate
from pyecore.utils import alias

from esdl import esdl
from esdl.utils.datetime_utils import parse_date


class ProfileType(str, Enum):
    """
    Different profile types
    """
    UNKNOWN = "UNKNOWN"
    # SINGLE_VALUE = "SINGLE_VALUE"
    # VALUE_LIST = "VALUE_LIST"
    DATETIME_LIST = "DATETIME_LIST"
    CSV = "CSV"
    EXCEL = "EXCEL"


class ProfileManager:
    """
    Represents a generic (set of) profile(s). Can be converted into ESDL profiles.
    """
    def __init__(self, source_profile=None):
        self.profile_header = None  # in case of CSV or Excel based profile data
        self.profile_data_list = []
        self.profile_type = ProfileType.UNKNOWN
        self.start_datetime = None
        self.end_datetime = None
        self.num_profile_items = 0

        if source_profile:
            self.convert(source_profile)

    def convert(self, source):
        """
        Copies all data from one profile instance to another. Is used when creating an InfluxDBProfile from an
        ExcelProfile.

        :param source: source profile where the data is copied from
        """
        self.profile_header = list(source.profile_header)
        self.profile_data_list = list(source.profile_data_list)
        self.profile_type = source.profile_type
        self.start_datetime = source.start_datetime
        self.end_datetime = source.end_datetime
        self.num_profile_items = source.num_profile_items

    def clear_profile(self):
        """
        Clears all profile information.
        """
        self.profile_header = None  # in case of CSV or Excel based profile data
        self.profile_data_list = []
        self.profile_type = ProfileType.UNKNOWN
        self.start_datetime = None
        self.end_datetime = None
        self.num_profile_items = 0

    def _check_data_format_and_size(self, profile_header, profile_data_list):
        num_columns = len(profile_header)
        if num_columns < 2:
            raise UnsupportedProfileInputDataException("profile_header should contain at least two items, " +
                                                       "a datetime column and at least one value column")

        for row in profile_data_list:
            if len(row) != num_columns:
                raise UnsupportedProfileInputDataException("one or more rows in profile_data_list don't contain the " +
                                                           "same amount of columns as the profile_header")
            if not isinstance(row[0], datetime):
                raise UnsupportedProfileInputDataException("first element in one or more rows in profile_data_list " +
                                                           "is not of type datetime")

    def set_profile(self, profile_header, profile_data_list, profile_type: ProfileType):
        """
        Sets the profile information

        :param profile_header: a list with column names, the first column is always the datetime column and also called
                               'datetime' (if you specify a different name, it's replaced by 'datetime')
        :param profile_data_list: a list of lists with values, the first item in the list must be of type datetime
        :param profile_type: the type of the profile
        """

        self._check_data_format_and_size(profile_header, profile_data_list)

        self.profile_header = profile_header
        if self.profile_header[0] != 'datetime':
            self.profile_header[0] = 'datetime'
        self.profile_data_list = profile_data_list
        self.profile_type = profile_type

        self.start_datetime = profile_data_list[0][0]
        self.end_datetime = profile_data_list[-1][0]
        self.num_profile_items = len(self.profile_data_list)

    def load_csv(self, file_path, encoding: str = "utf-8-sig"):
        """
        Reads profile data out of a CSV file.

        :param file_path: the path to the CSV file
        :param encoding: the used encoding in the CSV file (utf-8-sig by default)
        """
        self.clear_profile()
        self.profile_type = ProfileType.CSV

        try:
            csv_file = codecs.open(file_path, encoding=encoding)
            dialect = csv.Sniffer().sniff(csv_file.read(4096))
            csv_file.close()
            csv_file = codecs.open(file_path, encoding=encoding)
            reader = csv.reader(csv_file, dialect)
        except:
            # If format cannot be determined automatically, try ; as a default
            csv_file = codecs.open(file_path, encoding=encoding)
            reader = csv.reader(csv_file, delimiter=';')

        self.profile_header = next(reader)
        self.num_profile_items = 0

        previous_datetime = None
        for row in reader:
            try:
                dt = parse_date(row[0])
            except ValueError:  # ValueError: row in CSV doesn't start with recognizable datetime value, ignore row
                continue

            try:
                aware_dt = pytz.utc.localize(dt)  # Assume timezone is UTC if no TZ was given
            except ValueError:  # ValueError: No naive datetime (tzinfo is already set)
                aware_dt = dt
            row[0] = aware_dt
            for elem_idx in range(1, len(row)):
                try:
                    row[elem_idx] = locale.atof(row[elem_idx])      # Assume float values
                except Exception as e:
                    raise Exception(f"Cannot parse profile value in CSV ({row[elem_idx]})")

            if previous_datetime:
                if previous_datetime == aware_dt:
                    raise DuplicateValueInProfileException(
                        "CSV contains duplicate datetimes ({}). Check timezone and daylight saving".
                        format(aware_dt.strftime('%Y-%m-%dT%H:%M:%S%z'))
                    )
            previous_datetime = aware_dt

            if self.start_datetime is None:
                self.start_datetime = aware_dt

            self.profile_data_list.append(row)
            self.num_profile_items += 1

        self.end_datetime = self.profile_data_list[-1][0]
        csv_file.close()

    def parse_esdl(self, esdl_profile):
        """
        Parses an ESDL profile and loads the data

        :param esdl_profile: the ESDL profile object
        """
        if not isinstance(esdl_profile, esdl.TimeSeriesProfile) and not isinstance(esdl_profile, esdl.DateTimeProfile):
            raise Exception("Profile types other than esdl.TimeSeriesProfile or esdl.DateTimeProfile are not "
                            "supported yet. Use an InfluxDBProfile instance for esdl.InfluxDBProfile")

        self.clear_profile()
        self.profile_type = ProfileType.DATETIME_LIST

        name = esdl_profile.name if esdl_profile.name else 'NoName profile'
        self.profile_header = ['datetime', name]

        if isinstance(esdl_profile, esdl.TimeSeriesProfile):
            self.start_datetime = esdl_profile.startDateTime
            try:
                aware_dt = pytz.utc.localize(self.start_datetime)  # Assume timezone is UTC if no TZ was given
            except ValueError:  # ValueError: No naive datetime (tzinfo is already set)
                aware_dt = self.start_datetime

            for value in esdl_profile.values:
                self.profile_data_list.append([aware_dt, value])
                aware_dt = aware_dt + timedelta(seconds=esdl_profile.timestep)
                self.num_profile_items += 1
            self.end_datetime = self.profile_data_list[-1][0]

        if isinstance(esdl_profile, esdl.DateTimeProfile):
            if esdl_profile.element:
                for elem in esdl_profile.element:
                    dt = elem.from_
                    try:
                        aware_dt = pytz.utc.localize(dt)  # Assume timezone is UTC if no TZ was given
                    except ValueError:  # ValueError: No naive datetime (tzinfo is already set)
                        aware_dt = dt

                    if self.start_datetime is None:
                        self.start_datetime = aware_dt

                    self.profile_data_list.append([aware_dt, elem.value])
                    self.num_profile_items += 1
                self.end_datetime = self.profile_data_list[-1][0]
            else:
                raise Exception("Empty DateTimeProfile")

    def get_profile_name_index(self, profile_name):
        """
        Calculates the index of the column in the profile data array based on the column name

        :param profile_name: the name of the profile
        :return: the index of the column in the profile data array
        """
        if profile_name is not None:
            if profile_name in self.profile_header:
                return self.profile_header.index(profile_name)
            else:
                raise UnknownProfileNameException("Unknown profile name")
        else:
            return 1   # first element is datetime, second element is the default (or only) value

    def get_esdl_datetime_profile(self, profile_name=None):
        """
        Generates an esdl.DateTimeProfile from the loaded profile information

        :param profile_name: the name of the profile (only necessary when multiple columns have been loaded)
        :return: the esdl.DateTimeProfile
        """
        esdl.ProfileElement.from_.name = 'from'
        setattr(esdl.ProfileElement, 'from', esdl.ProfileElement.from_)

        if self.profile_type is ProfileType.UNKNOWN:
            raise NoProfileLoadedExecption("Cannot create DateTimeProfile, no profile information is loaded")
        profile_name_index = self.get_profile_name_index(profile_name)

        esdl_dt_profile = esdl.DateTimeProfile(id=str(uuid4()), name=profile_name)
        if profile_name:
            esdl_dt_profile.name = profile_name

        for profile_row in self.profile_data_list:
            elem = esdl.ProfileElement(
                from_=EDate.from_string(str(profile_row[0])),
                value=profile_row[profile_name_index]
            )
            esdl_dt_profile.element.append(elem)

        return esdl_dt_profile

    def get_esdl_timeseries_profile(self, profile_name=None):
        """
        Generates an esdl.TimeSeriesProfile from the loaded profile information

        :param profile_name: the name of the profile (only necessary when multiple columns have been loaded)
        :return: the esdl.TimeSeriesProfile
        """
        if self.profile_type is ProfileType.UNKNOWN:
            raise NoProfileLoadedExecption("Cannot create TimeSeriesProfile, no profile information is loaded")
        profile_name_index = self.get_profile_name_index(profile_name)

        esdl_ts_profile = esdl.TimeSeriesProfile(id=str(uuid4()), name=profile_name)
        if profile_name:
            esdl_ts_profile.name = profile_name
        esdl_ts_profile.startDateTime = EDate.from_string(str(self.start_datetime))

        if self.num_profile_items > 1:
            # Assume equidistant profile values for now
            esdl_ts_profile.timestep = int((self.profile_data_list[1][0]-self.profile_data_list[0][0]).total_seconds())
        else:
            esdl_ts_profile.timestep = 0

        for profile_row in self.profile_data_list:
            esdl_ts_profile.values.append(profile_row[profile_name_index])

        return esdl_ts_profile


class NoProfileLoadedExecption(Exception):
    """Thrown when no profile information is loaded"""
    pass


class UnknownProfileNameException(Exception):
    """Thrown when an unknown profile name is given"""
    pass


class DuplicateValueInProfileException(Exception):
    """Thrown when input data for a profile contains duplicate datetime values, probably due to a timezone error with
    daylight saving"""
    pass


class UnsupportedProfileInputDataException(Exception):
    """Thrown when input data for a profile is not recognized"""
    pass


# from distutils.core import setup
import itertools
import sys
import unittest

import versioneer
import setuptools

version = tuple(sys.version_info[:2])

if version < (3, 7):
    sys.exit('pyESDL requires at least Python >= 3.7')


# def my_test_suite():
#     test_loader = unittest.TestLoader()
#     test_suite = test_loader.discover('tests', pattern='*_test.py')
#     return test_suite


extras_require = {
    "profiles": ["influxdb==5.3.1", "openpyxl==3.1.2"],
    "geometry": ["shapely==2.0.1", "geojson==3.1.0", "pyproj==3.6.1"],
}
extras_require['all'] = list(itertools.chain.from_iterable(extras_require.values()))

setuptools.setup(
    name='pyESDL',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    url="https://energytransition.gitbook.io/esdl/",
    packages=setuptools.find_packages(),
    package_data={'': ['README.md', 'LICENSE.md']},
    include_package_data=True,
    license='Apache 2.0',
    long_description_content_type="text/markdown",
    long_description=open('README.md').read(),
    description="Python implementation of the Energy System Description Language (ESDL) for modelling energy systems",
    author='Ewoud Werkman',
    author_email='ewoud.werkman@tno.nl',
    python_requires='>=3.7',
    install_requires=[
        'pyecore==0.13.2'
    ],
    extras_require=extras_require,
    # test_suite='setup.my_test_suite',
    tests_require = ["pytest"]
)



#  This work is based on original code developed and copyrighted by TNO 2020.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO

"""
Support functions for managing EObjects
"""
from pyecore.ecore import EAttribute, EObject, EClass, EReference, EStructuralFeature
from pyecore.innerutils import InternalSet
from pyecore.valuecontainer import ECollection
import logging

logger = logging.getLogger(__name__)


# add support for shallow copying or cloning an object
# it copies the object's attributes (e.g. clone an object), does only shallow copying
def clone(self):
    """
    Shallow copying or cloning an object
    It only copies the object's attributes (e.g. clone an object)
    Usage object.clone() or copy.copy(object) (as _copy__() is also implemented)
    :param self:
    :return: A clone of the object
    """
    newone = type(self)()
    eclass = self.eClass
    for x in eclass.eAllStructuralFeatures():
        if isinstance(x, EAttribute):
            #logger.trace("clone: processing attribute {}".format(x.name))
            if x.many:
                eOrderedSet = newone.eGet(x.name)
                for v in self.eGet(x.name):
                    eOrderedSet.append(v)
            else:
                newone.eSet(x.name, self.eGet(x.name))
    return newone


def deepcopy(self, memo=None, target_es=None):
    """
    Deep copying an EObject.
    Does not work yet for copying references from other resources than this one.
    """
    #logger.debug("deepcopy: processing {}".format(self))
    first_call = False
    if memo is None:
        memo = dict()
        first_call = True
    if self in memo:
        return memo[self]

    copy: EObject = self.clone()
    #logger.debug("Shallow copy: {}".format(copy))
    eclass: EClass = self.eClass
    for x in eclass.eAllStructuralFeatures():
        if isinstance(x, EReference):
            #logger.debug("deepcopy: processing reference {}".format(x.name))
            ref: EReference = x
            value: EStructuralFeature = self.eGet(ref)
            if value is None:
                continue
            if ref.containment:
                if ref.many and isinstance(value, ECollection):
                    #clone all containment elements
                    eAbstractSet = copy.eGet(ref.name)
                    for ref_value in value:
                        duplicate = ref_value.__deepcopy__(memo, target_es)
                        eAbstractSet.append(duplicate)
                else:
                    copy.eSet(ref.name, value.__deepcopy__(memo, target_es))
            #else:
            #    # no containment relation, but a reference
            #    # this can only be done after a full copy
            #    pass
    # now copy should a full copy, but without cross references

    memo[self] = copy

    if first_call:
        #logger.debug("copying references")
        for k, v in memo.items():
            eclass: EClass = k.eClass
            for x in eclass.eAllStructuralFeatures():
                if isinstance(x, EReference):
                    #logger.debug("deepcopy: processing x-reference {}".format(x.name))
                    ref: EReference = x
                    orig_value: EStructuralFeature = k.eGet(ref)
                    if orig_value is None:
                        continue
                    if not ref.containment:
                        opposite = ref.eOpposite
                        if opposite and opposite.containment:
                            # do not handle eOpposite relations, they are handled automatically in pyEcore
                            continue
                        if x.many:
                            eAbstractSet = v.eGet(ref.name)
                            for orig_ref_value in orig_value:
                                copy_ref_value = None
                                try:
                                    copy_ref_value = memo[orig_ref_value]
                                except KeyError:
                                    try:
                                        if target_es:
                                            if hasattr(orig_ref_value, 'id'): # use id to find object in target_es
                                                r = target_es.eResource
                                                ref_id = orig_ref_value.id
                                                copy_ref_value = r.uuid_dict[ref_id]
                                                logger.warning(
                                                    f'Using reference of type {orig_ref_value.eClass.name} for reference {k.eClass.name}.{ref.name} in deepcopy memo, using {copy_ref_value}')
                                            else:
                                                logger.warning(
                                                    f'No id found, Cannot find reference of type {orig_ref_value.eClass.name} for reference {k.eClass.name}.{ref.name} in deepcopy memo, using original')
                                                copy_ref_value = orig_ref_value
                                        else:
                                            logger.warning(
                                                f'Cannot find reference of type {orig_ref_value.eClass.name} for reference {k.eClass.name}.{ref.name} in deepcopy memo, using original')
                                            copy_ref_value = orig_ref_value
                                    except Exception:
                                        logger.warning(f'Cannot find reference of type {orig_ref_value.eClass.name} for reference {k.eClass.name}.{ref.name} in deepcopy memo, using original')
                                        copy_ref_value = orig_ref_value
                                eAbstractSet.append(copy_ref_value)
                        else:
                            try:
                                copy_value = memo[orig_value]
                            except KeyError:
                                try:
                                    if target_es:
                                        if hasattr(orig_value, 'id'):  # use id to find object in target_es
                                            r = target_es.eResource
                                            ref_id = orig_value.id
                                            copy_value = r.uuid_dict[ref_id]
                                            logger.debug(
                                                f'Using reference of type {orig_value.eClass.name} for reference {k.eClass.name}.{ref.name} in deepcopy memo, using {copy_value}')
                                        else:
                                            logger.warning(
                                                f'Referenced object has no ID, Can\'t find reference for {orig_value.eClass.name} for reference {k.eClass.name}.{ref.name} in deepcopy memo or target_es, using original {orig_value}')
                                            copy_value = orig_value
                                    else:
                                        logger.warning(
                                            f'Can\'t find reference for {orig_value.eClass.name} for reference {k.eClass.name}.{ref.name} in deepcopy memo, using original')
                                        copy_value = orig_value
                                except Exception as e:
                                    logger.warning(f'{e}: Cannot find reference of type {orig_value.eClass.name} of reference {k.eClass.name}.{ref.name} in deepcopy memo, using original')
                                    copy_value = orig_value
                            v.eSet(ref.name, copy_value)
    copy._isset = InternalSet(self._isset)
    return copy



from pyecore.resources import global_registry
from .edd import getEClassifier, eClassifiers
from .edd import name, nsURI, nsPrefix, eClass
from .edd import EnergyDataDescription, Image

from pyecore.ecore import EObject

from . import edd

__all__ = ['EnergyDataDescription', 'Image']

eSubpackages = []
eSuperPackage = None
edd.eSubpackages = eSubpackages
edd.eSuperPackage = eSuperPackage

EnergyDataDescription.esdl.eType = EObject
EnergyDataDescription.image.eType = Image

otherClassifiers = []

for classif in otherClassifiers:
    eClassifiers[classif.name] = classif
    classif.ePackage = eClass

for classif in eClassifiers.values():
    eClass.eClassifiers.append(classif.eClass)

for subpack in eSubpackages:
    eClass.eSubpackages.append(subpack.eClass)

register_packages = [edd] + eSubpackages
for pack in register_packages:
    global_registry[pack.nsURI] = pack
